# (c) 2024 Niels Provos

# Standard library imports
import io
import base64
from pathlib import Path
from PIL import Image
import cv2

# Related third party imports
import numpy as np
from dash import dcc, html, ctx, no_update
from dash_extensions import EventListener
from dash.dependencies import Input, Output, State, ALL, ClientsideFunction
from dash.exceptions import PreventUpdate

# Local application/library specific imports
from . import constants as C
from .automatic1111 import make_models_request
from .comfyui import get_history, patch_inpainting_workflow
from .controller import AppState, CompositeMode
from .utils import to_image_url, find_square_bounding_box
from .inpainting import patch_image, create_inpainting_pipeline
from .segmentation import render_view, remove_mask_from_alpha
from .stabilityai import StabilityAI


def get_canvas_paint_events():
    props = [
        "type",
        "clientX",
        "clientY",
        "offsetX",
        "offsetY",
        "button",
        "altKey",
        "ctrlKey",
        "shiftKey",
    ]
    events = []
    for event in ["mousedown", "mouseup", "mouseout", "mouseenter"]:
        events.append({"event": event, "props": props})
    return events


def get_image_click_event():
    return {
        "event": "click",
        "props": [
            "type",
            "clientX",
            "clientY",
            "offsetX",
            "offsetY",
            "ctrlKey",
            "shiftKey",
        ],
    }


def make_input_image_container(
    upload_id: str = C.UPLOAD_IMAGE,
    image_id: str = C.IMAGE,
    event_id: str = "el",
    canvas_id: str = C.CANVAS,
    preview_canvas_id: str = C.PREVIEW_CANVAS,
    outer_class_name: str = "w-full col-span-2",
):

    return html.Div(
        [
            html.Label("Input Image", className="font-bold mb-2 ml-3"),
            dcc.Store(id=C.STORE_IGNORE),  # we don't read this data
            dcc.Store(id=C.STORE_CLICKED_POINT),
            dcc.Store(id=C.STORE_CLEAR_PREVIEW),
            dcc.Upload(
                id=upload_id,
                disabled=False,
                children=html.Div(
                    [
                        dcc.Loading(
                            id=C.LOADING_UPLOAD,
                            children=[],
                            fullscreen=False,  # Ensure not to use fullscreen
                            # Center in the Upload container
                            className="absolute top-1/2 left-1/2 transform -translate-x-1/2 -translate-y-1/2 z-20",
                            style={"color": "blue"},
                        ),
                        EventListener(
                            html.Img(
                                id=image_id,
                                className="absolute top-0 left-0 w-full h-full object-contain object-left-top z-0",
                            ),
                            events=[get_image_click_event()],
                            logging=True,
                            id=event_id,
                        ),
                        EventListener(
                            html.Canvas(
                                id=canvas_id,
                                className="absolute top-0 left-0 w-full h-full object-contain object-left-top opacity-50 z-10",
                            ),
                            id=C.CANVAS_PAINT,
                            events=get_canvas_paint_events(),
                            logging=False,
                        ),
                        html.Canvas(
                            id=preview_canvas_id,
                            className="absolute top-0 left-0 w-full h-full object-contain object-left-top opacity-50 z-20",
                        ),
                    ],
                    className="general-container relative flex h-full overflow-hidden",
                    style={"height": "67vh"},
                ),
                style={"height": "67vh"},
                className="w-full",
                disable_click=True,
                multiple=False,
            ),
            make_segmentation_tools_container(),
            make_inpainting_tools_container(),
        ],
        className=outer_class_name,
        id=C.CTR_INPUT_IMAGE,
    )


def make_inpainting_tools_container():
    return html.Div(
        [
            # Div for existing canvas tools
            html.Div(
                [
                    dcc.Store(id=C.CANVAS_DATA),  # saves to disk
                    dcc.Store(id=C.CANVAS_MASK_DATA),
                    # keeps track of the selected slice for Javascript
                    dcc.Store(id=C.STORE_SELECTED_SLICE),
                    html.Div(
                        [
                            html.Button(
                                "Clear",
                                id=C.BTN_CLEAR_CANVAS,
                                className="general-element mr-1",
                            ),
                            html.Button(
                                "Erase",
                                id=C.BTN_ERASE_MODE,
                                className="general-element mr-1",
                            ),
                            html.Button(
                                "Load",
                                id=C.BTN_LOAD_CANVAS,
                                className="general-element",
                            ),
                        ],
                        className="tools-container",
                    ),
                ],
                className="tools-backdrop items-center",
            ),
            # Div for navigation buttons arranged just in a single row
            html.Div(
                [
                    html.Div(
                        [
                            html.Button(
                                html.I(className="fa fa-magnifying-glass-minus"),
                                id=C.NAV_ZOOM_OUT,
                                className="nav-button mr-1",
                            ),
                            html.Button(
                                html.I(className="fa fa-arrow-up"),
                                id=C.NAV_UP,
                                className="nav-button mr-1",
                            ),
                            html.Button(
                                html.I(className="fa fa-arrow-left"),
                                id=C.NAV_LEFT,
                                className="nav-button mr-1",
                            ),
                            html.Button(
                                html.I(className="fa fa-circle"),
                                id=C.NAV_RESET,
                                className="nav-button mr-1",
                            ),
                            html.Button(
                                html.I(className="fa fa-arrow-right"),
                                id=C.NAV_RIGHT,
                                className="nav-button mr-1",
                            ),
                            html.Button(
                                html.I(className="fa fa-arrow-down"),
                                id=C.NAV_DOWN,
                                className="nav-button mr-1",
                            ),
                            html.Button(
                                html.I(className="fa fa-magnifying-glass-plus"),
                                id=C.NAV_ZOOM_IN,
                                className="nav-button",
                            ),
                        ],
                        className="tools-container",
                    ),
                ],
                className="tools-backdrop items-centergap-1",
            ),
        ],
        id=C.CTR_CANVAS_BUTTONS,
        className="flex gap-2",
    )


def make_segmentation_tools_container():
    return html.Div(
        [
            html.Div(
                [
                    html.Div(
                        [
                            html.Button(
                                [html.I(className="fa fa-clone")],
                                id=C.SEG_TOGGLE_CHECKERBOARD,
                                title="Toggle checkerboard background",
                                className="general-element color-not-selected mr-1",
                            ),
                            html.Button(
                                ["Invert ", html.I(className="fa fa-adjust")],
                                id=C.SEG_INVERT_MASK,
                                title="Invert the current mask",
                                className="general-element mr-1",
                            ),
                            html.Button(
                                ["Feather ", html.I(className="fa fa-wind")],
                                id=C.SEG_FEATHER_MASK,
                                title="Feather the current mask",
                                className="general-element mr-1",
                            ),
                            html.Button(
                                [
                                    "Multi ",
                                    html.I(className="fa fa-wand-magic-sparkles"),
                                ],
                                id=C.SEG_MULTI_POINT,
                                className="general-element color-not-selected mr-1",
                            ),
                            html.Button(
                                ["Commit ", html.I(className="fa fa-person-running")],
                                id=C.SEG_MULTI_COMMIT,
                                className="general-element mr-1",
                            ),
                        ],
                        className="flex justify-between",
                    )
                ],
                className="tools-backdrop items-center",
            )
        ],
        id=C.CTR_SEG_BUTTONS,
        className="inline-block",
    )


def make_tools_callbacks(app):
    @app.callback(
        Output(C.CANVAS, "className"),
        Output(C.IMAGE, "className"),
        Output(C.CTR_CANVAS_BUTTONS, "className"),
        Output(C.CTR_SEG_BUTTONS, "className"),
        Input({"type": "tab-content-main", "index": ALL}, "className"),
        State(C.CANVAS, "className"),
        State(C.IMAGE, "className"),
        State(C.CTR_CANVAS_BUTTONS, "className"),
        State(C.CTR_SEG_BUTTONS, "className"),
    )
    def update_events(
        tab_class_names,
        canvas_class_name,
        image_class_name,
        inpaint_btns_class,
        segment_btns_class,
    ):
        if tab_class_names is None:
            raise PreventUpdate()

        canvas_class_name = canvas_class_name.replace(" z-10", "").replace(" z-0", "")
        image_class_name = image_class_name.replace(" z-10", "").replace(" z-0", "")
        inpaint_btns_class = inpaint_btns_class.replace(" hidden", "")
        segment_btns_class = segment_btns_class.replace(" hidden", "")

        # tabs[1] == Segmentation tab
        # tabs[2] == Inpainting tab

        # we paint on the canvas only if the Inpainting tab is active
        if "hidden" not in tab_class_names[2]:
            canvas_class_name += " z-10"
            image_class_name += " z-0"
        else:
            canvas_class_name += " z-0"
            image_class_name += " z-10"
            inpaint_btns_class += " hidden"

        # we will show the segmentation tools only if the Segmentation tab is active
        if "hidden" in tab_class_names[1]:
            segment_btns_class += " hidden"

        return (
            canvas_class_name,
            image_class_name,
            inpaint_btns_class,
            segment_btns_class,
        )


def make_depth_map_container(depth_map_id: str = C.CTR_DEPTH_MAP):
    return html.Div(
        [
            html.Label("Depth Map", className="font-bold mb-2 ml-3"),
            html.Div(
                id=depth_map_id,
                className="general-container min-h-60",
            ),
            dcc.Loading(
                id=C.LOADING_DEPTHMAP,
                type="default",
                children=html.Div(id=C.DEPTHMAP_OUTPUT),
            ),
            html.Div(
                [
                    html.Div(
                        [
                            html.Label("Depth Module Algorithm"),
                            dcc.Dropdown(
                                id=C.DROPDOWN_DEPTH_MODEL,
                                options=[
                                    {"label": "MiDaS", "value": "midas"},
                                    {"label": "DINOv2", "value": "dinov2"},
                                ],
                                value="dinov2",
                                className="general-dropdown mt-2 mb-2",
                            ),
                        ],
                        className="w-full",
                    ),
                    html.Button(
                        html.Div(
                            [
                                html.Label("Regenerate Depth Map"),
                                html.I(className="fa-solid fa-image pl-1"),
                            ]
                        ),
                        id=C.BTN_GENERATE_DEPTHMAP,
                        className="general-element mt-2 mb-2",
                    ),
                ],
                className="w-full grid grid-cols-2 gap-2 p-2 items-end",
            ),
            html.Div(
                [
                    dcc.Interval(id=C.PROGRESS_INTERVAL, interval=500, n_intervals=0),
                    html.Div(id=C.CTR_PROGRESS_BAR, className="progress-bar"),
                ],
                className="p-2",
            ),
        ],
        className="w-full",
    )


def make_thresholds_container(thresholds_id: str = C.CTR_THRESHOLDS):
    return html.Div(
        [
            html.Label("Thresholds", className="font-bold mb-2 ml-3"),
            html.Div(
                id=thresholds_id,
                className="general-border min-h-8 w-full flex-auto grow",
            ),
        ],
        className="w-full mb-2",
    )


def make_slice_generation_container():
    return html.Div(
        [
            dcc.Store(id=C.STORE_GENERATE_SLICE),
            dcc.Store(id=C.STORE_UPDATE_SLICE),
            dcc.Download(id=C.DOWNLOAD_IMAGE),
            html.Div(
                [
                    html.Div(
                        [
                            make_thresholds_container(thresholds_id=C.CTR_THRESHOLDS),
                        ],
                        className="w-full col-span-3",
                    ),
                    html.Div(
                        [
                            html.Label("Actions", className="font-bold mb-1 ml-3"),
                            html.Div(
                                [
                                    html.Button(
                                        html.Div(
                                            [
                                                html.Label("Generate"),
                                                html.I(
                                                    className="fa-solid fa-images pl-1"
                                                ),
                                            ]
                                        ),
                                        id=C.BTN_GENERATE_SLICE,
                                        title="Generate image slices from the input image using the depth map",
                                        className="w-full general-element mb-1 mr-2",
                                    ),
                                    html.Button(
                                        html.Div(
                                            [
                                                html.Label("Balance"),
                                                html.I(
                                                    className="fa-solid fa-arrows-left-right pl-1"
                                                ),
                                            ]
                                        ),
                                        id=C.BTN_BALANCE_SLICE,
                                        title="Rebalances the depths of the image slices evenly",
                                        className="w-full general-element mb-1",
                                    ),
                                    html.Button(
                                        html.Div(
                                            [
                                                html.Label("Create"),
                                                html.I(
                                                    className="fa-solid fa-square-plus pl-1"
                                                ),
                                            ]
                                        ),
                                        id=C.BTN_CREATE_SLICE,
                                        title="Creates a slice from the current mask",
                                        className="w-full general-element mb-1",
                                    ),
                                    html.Button(
                                        html.Div(
                                            [
                                                html.Label("Delete"),
                                                html.I(
                                                    className="fa-solid fa-trash-can pl-1"
                                                ),
                                            ]
                                        ),
                                        id=C.BTN_DELETE_SLICE,
                                        title="Deletes the currently selected slice",
                                        className="w-full general-element mb-1",
                                    ),
                                    html.Button(
                                        html.Div(
                                            [
                                                html.Label("Add"),
                                                html.I(
                                                    className="fa-solid fa-brush pl-1"
                                                ),
                                            ]
                                        ),
                                        id=C.BTN_ADD_SLICE,
                                        title="Adds the current mask to the selected slice",
                                        className="w-full general-element mb-1",
                                    ),
                                    html.Button(
                                        html.Div(
                                            [
                                                html.Label("Remove"),
                                                html.I(
                                                    className="fa-solid fa-eraser pl-1"
                                                ),
                                            ]
                                        ),
                                        id=C.BTN_REMOVE_SLICE,
                                        title="Removes the current mask from the selected slice",
                                        className="w-full general-element mb-1",
                                    ),
                                    html.Button(
                                        html.Div(
                                            [
                                                html.Label("Copy"),
                                                html.I(
                                                    className="fa-solid fa-brush pl-1"
                                                ),
                                            ]
                                        ),
                                        id=C.BTN_COPY_SLICE,
                                        title="Copies the current mask to the clipboard",
                                        className="w-full general-element mb-1",
                                    ),
                                    html.Button(
                                        html.Div(
                                            [
                                                html.Label("Paste"),
                                                html.I(
                                                    className="fa-solid fa-eraser pl-1"
                                                ),
                                            ]
                                        ),
                                        id=C.BTN_PASTE_SLICE,
                                        title="Copies the clipboard to the selected slice",
                                        className="w-full general-element mb-1",
                                    ),
                                ],
                                className="grid grid-cols-2 gap-2 gap-2 p-2",
                            ),
                        ],
                        className="w-full h-full col-span-2",
                    ),
                ],
                className="grid grid-cols-5 gap-4 p-2",
            ),
            dcc.Loading(
                id=C.LOADING_GENERATE_SLICE, children=html.Div(id="gen-slice-output")
            ),
            html.Div(
                id=C.CTR_SLICE_IMAGES,
                className="general-border min-h-8 w-full grid grid-cols-3 gap-1 overflow-auto",
            ),
        ],
        className="w-full overflow-auto",
        style={"height": "69vh"},
    )


def make_inpainting_container():
    return html.Div(
        [
            dcc.Store(id=C.STORE_INPAINTING),
            html.Div(
                [
                    html.Label("Positive Prompt"),
                    dcc.Textarea(
                        id=C.TEXT_POSITIVE_PROMPT,
                        placeholder="Enter a positive generative AI prompt...",
                        className="w-full light-border",
                        style={"height": "100px"},
                    ),
                ]
            ),
            html.Div(
                [
                    html.Label("Negative Prompt"),
                    dcc.Textarea(
                        id=C.TEXT_NEGATIVE_PROMPT,
                        placeholder="Enter a negative prompt...",
                        className="w-full light-border",
                    ),
                ]
            ),
            html.Div(
                [
                    html.Div(
                        [
                            html.Label("Strength"),
                            dcc.Slider(
                                id=C.SLIDER_INPAINT_STRENGTH,
                                min=0,
                                max=1,
                                step=0.01,
                                value=0.8,
                                marks=None,
                                tooltip={"placement": "bottom", "always_visible": True},
                            ),
                        ],
                        className="w-full p-2",
                    ),
                    html.Div(
                        [
                            html.Label("Guidance Scale"),
                            dcc.Slider(
                                id=C.SLIDER_INPAINT_GUIDANCE,
                                min=1,
                                max=15,
                                step=0.25,
                                value=7.5,
                                marks=None,
                                tooltip={"placement": "bottom", "always_visible": True},
                            ),
                        ],
                        className="w-full p-2",
                    ),
                ],
                className="flex w-full",
            ),
            dcc.Checklist(
                id=C.CHECKLIST_REGION_OF_INTEREST,
                options=[
                    {
                        "label": html.Span(
                            "Crop to region of interest", className="p-2"
                        ),
                        "value": "crop",
                    }
                ],
                value=["crop"],
                className="p-2",
            ),
            html.Button(
                html.Div(
                    [
                        html.Label("Generate"),
                        html.I(className="fa-solid fa-paint-brush pl-1"),
                    ]
                ),
                id=C.BTN_GENERATE_INPAINTING,
                title="Inpaint the areas of the selected image based on the painted mask",
                className="general-element mb-2 mt-3 mr-2",
            ),
            html.Button(
                html.Div(
                    [html.Label("Fill"), html.I(className="fa-solid fa-fill pl-1")]
                ),
                id=C.BTN_FILL_INPAINTING,
                title="Fill all empty areas of the selected image via inpainting",
                className="general-element mb-2 mt-3 mr-2",
            ),
            html.Button(
                html.Div(
                    [
                        html.Label("Enhance"),
                        html.I(className="fa-solid fa-wand-magic-sparkles pl-1"),
                    ]
                ),
                id=C.BTN_ENHANCE,
                title="Use inpainting to enhance the selected image via up and downscaling",
                className="general-element mb-2 mt-3 mr-2",
            ),
            html.Button(
                html.Div(
                    [
                        html.Label("Erase"),
                        html.I(className="fa-solid fa-trash-can pl-1"),
                    ]
                ),
                id=C.BTN_ERASE_INPAINTING,
                className="general-element mb-2 mt-3",
                title="Clear the painted areas from the selected image",
            ),
            dcc.Loading(
                id=C.LOADING_GENERATE_INPAINTING,
                children=html.Div(id=C.CTR_INPAINTING_OUTPUT),
            ),
            html.Div(
                id=C.CTR_INPAINTING_IMAGES,
                children=[
                    html.Div(
                        id=C.CTR_INPAINTING_DISPLAY, className="grid grid-cols-3 gap-2"
                    ),
                ],
                className="general-border w-full min-h-8",
            ),
            html.Button(
                "Apply Selected Image",
                id=C.BTN_APPLY_INPAINTING,
                className="color-is-selected p-2 rounded-md mt-2",
                disabled=True,
            ),
        ],
        className="w-full",
        id=C.CTR_INPAINTING_COLUNM,
    )


def make_inpainting_container_callbacks(app):
    @app.callback(
        Output(C.BTN_APPLY_INPAINTING, "disabled"),
        Input(C.CTR_INPAINTING_DISPLAY, "children"),
    )
    def enable_apply_inpainting_button(children):
        return False if children else True

    @app.callback(
        Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
        Output(C.LOGS_DATA, "data", allow_duplicate=True),
        Input(C.BTN_ERASE_INPAINTING, "n_clicks"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State(C.LOGS_DATA, "data"),
        running=[(Output(C.BTN_ERASE_INPAINTING, "disabled"), True, False)],
        prevent_initial_call=True,
    )
    def erase_inpainting(n_clicks, filename, logs):
        if n_clicks is None or filename is None:
            raise PreventUpdate()

        state = AppState.from_cache(filename)
        if state.selected_slice is None:
            logs.append("No slice selected")
            return no_update, logs

        index = state.selected_slice
        mask_filename = state.mask_filename(index)
        if not Path(mask_filename).exists():
            logs.append(f"No mask found for slice {index}")
            return no_update, logs
        mask = Image.open(mask_filename).convert("L")
        mask = np.array(mask)

        final_mask = remove_mask_from_alpha(state.image_slices[index].image, mask)
        state.image_slices[index].image[:, :, 3] = final_mask
        state.image_slices[index].new_version()

        state.to_file(
            state.filename,
            save_image_slices=False,
            save_depth_map=False,
            save_input_image=False,
        )
        logs.append(f"Inpainting erased for slice {index}")

        logs.append(f"Inpainting erased for slice {index}")

        return True, logs

    @app.callback(
        Output(C.CTR_INPAINTING_DISPLAY, "children"),
        Output(C.LOADING_GENERATE_INPAINTING, "children"),
        Input(C.BTN_GENERATE_INPAINTING, "n_clicks"),
        Input(C.BTN_FILL_INPAINTING, "n_clicks"),
        Input(C.BTN_ENHANCE, "n_clicks"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State(C.DROPDOWN_INPAINT_MODEL, "value"),
        State(C.UPLOAD_COMFYUI_WORKFLOW, "contents"),
        State(C.TEXT_POSITIVE_PROMPT, "value"),
        State(C.TEXT_NEGATIVE_PROMPT, "value"),
        State(C.SLIDER_INPAINT_STRENGTH, "value"),
        State(C.SLIDER_INPAINT_GUIDANCE, "value"),
        State(C.SLIDER_MASK_PADDING, "value"),
        State(C.SLIDER_MASK_BLUR, "value"),
        running=[
            (Output(C.BTN_GENERATE_INPAINTING, "disabled"), True, False),
            (Output(C.BTN_FILL_INPAINTING, "disabled"), True, False),
            (Output(C.BTN_ENHANCE, "disabled"), True, False),
        ],
        prevent_initial_call=True,
    )
    def update_inpainting_image_display(
        n_clicks_one,
        n_clicks_two,
        n_clicks_three,
        filename,
        model,
        workflow,
        positive_prompt,
        negative_prompt,
        strength,
        guidance_scale,
        padding,
        blur,
    ):
        if n_clicks_one is None and n_clicks_two is None and n_clicks_three is None:
            raise PreventUpdate()

        if filename is None:
            raise PreventUpdate()

        state = AppState.from_cache(filename)
        if state.selected_slice is None:
            raise PreventUpdate()  # XXX - write controller logic to clear this on image changes

        index = state.selected_slice
        # An empty prompt is OK.
        if positive_prompt is None:
            positive_prompt = ""
        if negative_prompt is None:
            negative_prompt = ""

        state.image_slices[state.selected_slice].positive_prompt = positive_prompt
        state.image_slices[state.selected_slice].negative_prompt = negative_prompt
        state.to_file(
            state.filename,
            save_image_slices=False,
            save_depth_map=False,
            save_input_image=False,
        )

        tid = ctx.triggered_id

        image = state.image_slices[index].image

        pipeline = create_inpainting_pipeline(model, workflow, state)

        if tid == C.BTN_GENERATE_INPAINTING or tid == C.BTN_FILL_INPAINTING:
            if tid == C.BTN_GENERATE_INPAINTING:
                mask_filename = state.mask_filename(index)
                if not Path(mask_filename).exists():
                    raise PreventUpdate()

                mask = Image.open(mask_filename).convert("L")
                mask = np.array(mask)
            else:
                # we'll fill everything that does not have an alpha
                mask = 255 - image[:, :, 3]

            def execute(input_image):
                return pipeline.inpaint(
                    positive_prompt,
                    negative_prompt,
                    input_image,
                    mask,
                    strength=strength,
                    guidance_scale=guidance_scale,
                    blur_radius=blur,
                    padding=padding,
                    crop=True,
                )

            # patch the image
            image = patch_image(image, mask)
            num_images = 3
        else:
            assert tid == C.BTN_ENHANCE

            def upscale_image(input_image):
                upscaled_image = state.upscale_image(
                    input_image, prompt=positive_prompt, negative_prompt=negative_prompt
                )
                upscaled_image = upscaled_image.resize(
                    (input_image.shape[1], input_image.shape[0]), Image.LANCZOS
                )
                upscaled_image = np.array(upscaled_image)
                upscaled_image[:, :, 3] = input_image[:, :, 3]
                return upscaled_image

            def execute(input_image):
                return upscale_image(input_image)

            num_images = 2

        images = []
        for i in range(num_images):
            new_image = execute(image)
            images.append(new_image)

        children = []
        for i, new_image in enumerate(images):
            children.append(
                html.Img(
                    src=to_image_url(new_image),
                    className="w-full h-full object-contain",
                    id={"type": C.ID_INPAINTING_IMAGE, "index": i},
                )
            )
        return children, []

    @app.callback(
        Output(C.IMAGE, "src", allow_duplicate=True),
        Output({"type": C.ID_INPAINTING_IMAGE, "index": ALL}, "className"),
        Output(C.LOADING_UPLOAD, "children", allow_duplicate=True),
        Input({"type": C.ID_INPAINTING_IMAGE, "index": ALL}, "n_clicks"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State({"type": C.ID_INPAINTING_IMAGE, "index": ALL}, "src"),
        State({"type": C.ID_INPAINTING_IMAGE, "index": ALL}, "className"),
        prevent_initial_call=True,
    )
    def select_inpainting_image(n_clicks, filename, images, classnames):
        if n_clicks is None or filename is None:
            raise PreventUpdate()

        index = ctx.triggered_id["index"]
        if n_clicks[index] is None:
            raise PreventUpdate()

        state = AppState.from_cache(filename)
        if state.selected_slice is None:
            raise PreventUpdate()

        state.selected_inpainting = index

        print(f"Applying inpainting image {index}")

        # give a visual highlight on the selected children
        return_image = no_update
        new_classnames = []
        selected_background = " color-is-selected-light"
        for i, classname in enumerate(classnames):
            already_selected = selected_background in classname
            classname = classname.replace(selected_background, "")
            if i == index:
                if not already_selected:
                    classname += selected_background
                    return_image = images[i]
                else:
                    mode = (
                        CompositeMode.CHECKERBOARD
                        if state.use_checkerboard
                        else CompositeMode.GRAYSCALE
                    )
                    return_image = state.serve_slice_image_composed(
                        state.selected_slice, mode
                    )
            new_classnames.append(classname)

        return return_image, new_classnames, ""

    @app.callback(
        Output(C.STORE_INPAINTING, "data"),
        Output(C.LOGS_DATA, "data", allow_duplicate=True),
        Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
        Input(C.BTN_APPLY_INPAINTING, "n_clicks"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State({"type": C.ID_INPAINTING_IMAGE, "index": ALL}, "src"),
        State(C.LOGS_DATA, "data"),
        running=[(Output(C.BTN_APPLY_INPAINTING, "disabled"), True, False)],
        prevent_initial_call=True,
    )
    def apply_inpainting(n_clicks, filename, inpainted_images, logs):
        if n_clicks is None or filename is None:
            raise PreventUpdate()

        state = AppState.from_cache(filename)
        if state.selected_inpainting is None:
            raise PreventUpdate()

        index = state.selected_slice
        new_image_data = inpainted_images[state.selected_inpainting]

        new_image = Image.open(
            io.BytesIO(base64.b64decode(new_image_data.split(",")[1]))
        )

        image_filename = state.image_slices[index].new_version(np.array(new_image))
        state.to_file(
            state.filename,
            save_image_slices=False,
            save_depth_map=False,
            save_input_image=False,
        )

        logs.append(
            f"Inpainting applied to slice {index} with new image {image_filename}"
        )

        return True, logs, True

    # this is called when the selected slice changes
    @app.callback(
        Output(C.TEXT_POSITIVE_PROMPT, "disabled"),
        Output(C.TEXT_NEGATIVE_PROMPT, "disabled"),
        Output(C.BTN_GENERATE_INPAINTING, "disabled"),
        Output(C.BTN_FILL_INPAINTING, "disabled"),
        Output(C.BTN_ENHANCE, "disabled"),
        Output(C.BTN_ERASE_INPAINTING, "disabled"),
        Output(C.STORE_SELECTED_SLICE, "data"),
        Input(C.STORE_INPAINTING, "data"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        prevent_initial_call=True,
    )
    def react_selected_slice_change(ignore, filename):
        if filename is None:
            return True, True, True, True, True, True, None

        state = AppState.from_cache(filename)
        if state.selected_slice is None:
            return True, True, True, True, True, True, None

        return False, False, False, False, False, False, state.selected_slice

    return update_inpainting_image_display


def make_configuration_callbacks(app):
    success_class = " color-is-selected-light"
    failure_class = " failure-color"

    @app.callback(
        Output(C.UPLOAD_COMFYUI_WORKFLOW, "contents"),
        Output(C.UPLOAD_COMFYUI_WORKFLOW, "children"),
        Output(C.LOGS_DATA, "data", allow_duplicate=True),
        Input(C.UPLOAD_COMFYUI_WORKFLOW, "contents"),
        State(C.UPLOAD_COMFYUI_WORKFLOW, "filename"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State(C.LOGS_DATA, "data"),
        prevent_initial_call=True,
    )
    def validate_workflow(contents, upload_name, filename, logs):
        if contents is None:
            raise PreventUpdate()

        # remove the data URL prefix
        contents = contents.split(",")[1]
        contents = base64.b64decode(contents)

        try:
            patch_inpainting_workflow(contents, C.IMAGE, "mask", "positive", "negative")
            logs.append("ComfyUI workflow validated")
        except Exception as e:
            logs.append(f"ComfyUI workflow validation failed: {str(e)}")
            return (
                "",
                [
                    "Drag and Drop or ",
                    html.I(className="fa-solid fa-upload"),
                    " to upload",
                ],
                logs,
            )

        if filename is not None:
            state = AppState.from_cache(filename)
            workflow_path = state.workflow_path()

            if not workflow_path.exists() or workflow_path.read_bytes() != contents:
                with open(workflow_path, "wb") as f:
                    f.write(contents)
                state.to_file(
                    state.filename,
                    save_image_slices=False,
                    save_depth_map=False,
                    save_input_image=False,
                )

        return no_update, upload_name, logs

    # Stability AI API for inpainting does not support mask blurring
    @app.callback(
        Output(C.SLIDER_INPAINT_GUIDANCE, "disabled"),
        Input(C.DROPDOWN_INPAINT_MODEL, "value"),
        prevent_initial_call=True,
    )
    def toggle_blur_slider(value):
        if value == "stabilityai":
            return True
        return False

    @app.callback(
        Output(C.CTR_AUTOMATIC_CONFIG, "className"),
        Output(C.CTR_COMFYUI_WORKFLOW, "className"),
        Input(C.DROPDOWN_INPAINT_MODEL, "value"),
        State(C.CTR_AUTOMATIC_CONFIG, "className"),
        State(C.CTR_COMFYUI_WORKFLOW, "className"),
    )
    def toggle_automatic_config(value, class_name_server, class_name_workflow):
        class_name_server = class_name_server.replace(" hidden", "")
        if value != "automatic1111" and value != "comfyui":
            class_name_server += " hidden"

        class_name_workflow = class_name_workflow.replace(" hidden", "")
        if value != "comfyui":
            class_name_workflow += " hidden"
        return class_name_server, class_name_workflow

    @app.callback(
        Output(C.INPUT_EXTERNAL_SERVER, "className", allow_duplicate=True),
        Input(C.INPUT_EXTERNAL_SERVER, "value"),
        State(C.INPUT_EXTERNAL_SERVER, "className"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        prevent_initial_call=True,
    )
    def reset_external_server_address(value, class_name, filename):
        if filename is not None:
            state = AppState.from_cache(filename)
            if state.server_address == value:
                raise PreventUpdate()
            state.server_address = value
            state.to_file(
                state.filename,
                save_image_slices=False,
                save_depth_map=False,
                save_input_image=False,
            )

        return class_name.replace(success_class, "").replace(failure_class, "")

    @app.callback(
        Output(C.LOGS_DATA, "data", allow_duplicate=True),
        Output(C.INPUT_EXTERNAL_SERVER, "className"),
        Input(C.BTN_EXTERNAL_TEST_CONNECTION, "n_clicks"),
        State(C.INPUT_EXTERNAL_SERVER, "value"),
        State(C.INPUT_EXTERNAL_SERVER, "className"),
        State(C.DROPDOWN_INPAINT_MODEL, "value"),
        State(C.LOGS_DATA, "data"),
        prevent_initial_call=True,
    )
    def test_external_connection(n_clicks, server_address, class_name, model, logs):
        if n_clicks is None:
            raise PreventUpdate()

        class_name = class_name.replace(success_class, "").replace(failure_class, "")

        success = False
        try:
            if model == "automatic1111":
                data = make_models_request(server_address)
            else:
                data = get_history(server_address, "test")
            if data is not None:
                logs.append(f"Connection to {model} successful: {data}")
                success = True
            else:
                logs.append(f"Connection to {model} failed")
        except Exception as e:
            logs.append(f"Connection to {model} failed: {str(e)}")

        class_name += success_class if success else failure_class

        return logs, class_name

    # Stability AI related callbacks
    @app.callback(
        Output(C.CTR_API_KEY, "className"),
        Input(C.DROPDOWN_INPAINT_MODEL, "value"),
        State(C.CTR_API_KEY, "className"),
    )
    def toggle_stabilityai_config(value, class_name):
        class_name = class_name.replace(" hidden", "")
        if value != "stabilityai":
            class_name += " hidden"
        return class_name

    @app.callback(
        Output(C.INPUT_API_KEY, "className", allow_duplicate=True),
        Input(C.INPUT_API_KEY, "value"),
        State(C.INPUT_API_KEY, "className"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        prevent_initial_call=True,
    )
    def reset_external_api_key(value, class_name, filename):
        if filename is not None:
            state = AppState.from_cache(filename)
            if state.api_key == value:
                raise PreventUpdate()
            state.api_key = value
            state.to_file(
                state.filename,
                save_image_slices=False,
                save_depth_map=False,
                save_input_image=False,
            )

        return class_name.replace(success_class, "").replace(failure_class, "")

    @app.callback(
        Output(C.LOGS_DATA, "data", allow_duplicate=True),
        Output(C.INPUT_API_KEY, "className"),
        Input(C.BTN_VALIDATE_API_KEY, "n_clicks"),
        State(C.INPUT_API_KEY, "value"),
        State(C.INPUT_API_KEY, "className"),
        State(C.DROPDOWN_INPAINT_MODEL, "value"),
        State(C.LOGS_DATA, "data"),
        prevent_initial_call=True,
    )
    def test_api_key(n_clicks, api_key, class_name, model, logs):
        if n_clicks is None:
            raise PreventUpdate()

        class_name = class_name.replace(success_class, "").replace(failure_class, "")

        success = False
        if api_key.startswith("sk-") and len(api_key) > 12:
            try:
                inpaint_model = StabilityAI(api_key)
                success, credits = inpaint_model.validate_key()
                if success:
                    logs.append(
                        f"Connection to {model} successful: you have {credits:0.2f} remaining credits"
                    )
                    success = True
                else:
                    logs.append(f"Connection to {model} failed")
            except Exception as e:
                logs.append(f"Connection to {model} failed: {str(e)}")
        else:
            logs.append(f"Invalid API key format")

        class_name += success_class if success else failure_class

        return logs, class_name


def make_configuration_container():
    return make_label_container("Configuration", make_configuration_div())


def make_configuration_div():
    return html.Div(
        [
            html.Div(
                [
                    html.Label("Number of Slices"),
                    dcc.Slider(
                        id=C.SLIDER_NUM_SLICES,
                        min=2,
                        max=10,
                        step=1,
                        value=3,
                        marks={i: str(i) for i in range(2, 11)},
                    ),
                ],
                className="w-full",
            ),
            html.Div(
                [
                    html.Label("Inpainting Model"),
                    dcc.Dropdown(
                        id=C.DROPDOWN_INPAINT_MODEL,
                        options=[
                            {
                                "label": "Kadinksy",
                                "value": "kandinsky-community/kandinsky-2-2-decoder-inpaint",
                            },
                            {
                                "label": "SD 1.5",
                                "value": "unwayml/stable-diffusion-v1-5",
                            },
                            {
                                "label": "SD XL 1.0",
                                "value": "diffusers/stable-diffusion-xl-1.0-inpainting-0.1",
                            },
                            {
                                "label": "StableDiffusion3",
                                "value": "stabilityai/stable-diffusion-3-medium-diffusers",
                            },
                            {"label": "Automatic1111", "value": "automatic1111"},
                            {"label": "ComfyUI", "value": "comfyui"},
                            {"label": "StabilityAI", "value": "stabilityai"},
                        ],
                        value="diffusers/stable-diffusion-xl-1.0-inpainting-0.1",
                        className="general-dropdown",
                    ),
                ],
                className="w-full",
            ),
            html.Div(
                [
                    html.Label("A1111/ComfyUI Server Address"),
                    html.Div(
                        [
                            dcc.Input(
                                id=C.INPUT_EXTERNAL_SERVER,
                                value="localhost:7860",
                                type="text",
                                debounce=True,
                                className="light-border flex-grow",
                            ),
                            html.Button(
                                html.Div(
                                    [
                                        html.Label("Test Connection"),
                                        html.I(
                                            className="fa-solid fa-network-wired pl-1"
                                        ),
                                    ]
                                ),
                                id=C.BTN_EXTERNAL_TEST_CONNECTION,
                                className="general-element mb-2 ml-2",
                            ),
                        ],
                        # Set the container to display flex for a row layout
                        className="flex flex-row items-center w-full",
                    ),
                    html.Div(
                        [
                            html.Label("ComfyUI Workflow"),
                            dcc.Upload(
                                children=[
                                    "Drag and Drop or ",
                                    html.I(className="fa-solid fa-upload"),
                                    " to upload",
                                ],
                                className="light-border flex-grow",
                                id=C.UPLOAD_COMFYUI_WORKFLOW,
                            ),
                        ],
                        id=C.CTR_COMFYUI_WORKFLOW,
                        className="w-full",
                    ),
                ],
                id=C.CTR_AUTOMATIC_CONFIG,
                className="w-full",
            ),
            html.Div(
                [
                    html.Label("StabilityAI API Key"),
                    html.Div(
                        [
                            dcc.Input(
                                id=C.INPUT_API_KEY,
                                value="",
                                type="password",
                                debounce=True,
                                className="light-border flex-grow",
                            ),
                            html.Button(
                                html.Div(
                                    [
                                        html.Label("Test API Key"),
                                        html.I(
                                            className="fa-solid fa-network-wired pl-1"
                                        ),
                                    ]
                                ),
                                id=C.BTN_VALIDATE_API_KEY,
                                className="general-element mb-2 ml-2",
                            ),
                        ],
                        # Set the container to display flex for a row layout
                        className="flex flex-row items-center w-full",
                    ),
                ],
                id=C.CTR_API_KEY,
                className="w-full",
            ),
            html.Div(
                [
                    html.Label("Inpainting Parameters"),
                    html.Div(
                        [
                            html.Label("Mask Padding"),
                            dcc.Slider(
                                id=C.SLIDER_MASK_PADDING,
                                min=0,
                                max=200,
                                step=10,
                                value=50,
                                marks={i * 10: str(i * 10) for i in range(21)},
                            ),
                            html.Label("Mask Blur"),
                            dcc.Slider(
                                id=C.SLIDER_MASK_BLUR,
                                min=0,
                                max=200,
                                step=10,
                                value=50,
                                marks={i * 10: str(i * 10) for i in range(21)},
                            ),
                        ],
                        className="general-border w-full min-h-8",
                    ),
                ],
                className="w-full",
            ),
            html.Div(
                [
                    html.Label("Export/Import State"),
                    html.Div(
                        [
                            dcc.Upload(
                                html.Button(
                                    html.Div(
                                        [
                                            html.Label("Load State"),
                                            html.I(className="fa-solid fa-upload pl-1"),
                                        ]
                                    ),
                                    className="general-element mb-2",
                                ),
                                id=C.UPLOAD_STATE,
                                multiple=False,
                            ),
                            html.Button(
                                html.Div(
                                    [
                                        html.Label("Save State"),
                                        html.I(className="fa-solid fa-download pl-1"),
                                    ],
                                    className="general-element mb-2",
                                ),
                                id=C.BTN_SAVE_STATE,
                            ),
                        ],
                        className="flex flex-row gap-4",
                    ),
                ],
                className="w-full mt-2",
            ),
        ]
    )


def make_3d_export_div():
    return html.Div(
        [
            dcc.Loading(id=C.LOADING_GLTF, children=html.Div(id=C.CTR_GLTF_OUTPUT)),
            html.Button(
                html.Div(
                    [
                        html.Label("Create glTF Scene"),
                        html.I(className="fa-solid fa-cube pl-1"),
                    ]
                ),
                id=C.BTN_GLTF_CREATE,
                className="general-element mb-2 mr-2",
            ),
            html.Button(
                html.Div(
                    [
                        html.Label("Export glTF Scene"),
                        html.I(className="fa-solid fa-download pl-1"),
                    ]
                ),
                id=C.BTN_GLTF_EXPORT,
                className="general-element mb-2 mr-2",
            ),
            html.Button(
                html.Div(
                    [
                        html.Label("Upscale Textures"),
                        html.I(className="fa-solid fa-maximize pl-1"),
                    ]
                ),
                id=C.BTN_UPSCALE_TEXTURES,
                className="general-element mb-2",
            ),
            dcc.Checklist(
                id=C.CHECKLIST_DOF,
                options=[
                    {
                        "label": html.Span(
                            "Support Depth of Field Effect", className="p-2"
                        ),
                        "value": "dof",
                    }
                ],
                value=[""],
                className="p-2",
            ),
            make_slider(C.SLIDER_CAMERA_DISTANCE, "Camera Distance", 0, 500, 1, 100),
            make_slider(C.SLIDER_MAX_DISTANCE, "Max Distance", 0, 1000, 1, 200),
            make_slider(C.SLIDER_FOCAL_LENGTH, "Focal Length", 0, 500, 1, 100),
            html.Label("Mesh Displacement"),
            dcc.Slider(
                id=C.SLIDER_DISPLACEMENT,
                min=0,
                max=150,
                step=5,
                value=0,
                marks={i * 10: str(i * 10) for i in range(16)},
            ),
            dcc.Download(id=C.DOWNLOAD_GLTF),
        ],
        className="general-border min-h-8 w-full flex-auto grow mb-2",
    )


def make_animation_export_div():
    return html.Div(
        [
            html.Button(
                html.Div(
                    [
                        html.Label("Export Animation"),
                        html.I(className="fa-solid fa-download pl-1"),
                    ]
                ),
                id=C.BTN_EXPORT_ANIMATION,
                className="general-element mb-2",
            ),
            dcc.Loading(
                id=C.LOADING_ANIMATION, children=html.Div(id=C.ANIMATION_OUTPUT)
            ),
            make_slider(C.SLIDER_NUM_FRAMES, "Number of Frames", 0, 300, 1, 100),
            dcc.Download(id=C.DOWNLOAD_ANIMATION),
        ],
        className="min-h-8 w-full flex-auto grow general-border mb-2",
    )


def make_slider(
    slider_id: str, label: str, min_value: int, max_value: int, step: int, value: int
):
    return html.Div(
        [
            html.P(label),
            dcc.Slider(
                id=slider_id,
                min=min_value,
                max=max_value,
                step=step,
                value=value,
                marks=None,
                tooltip={"placement": "bottom", "always_visible": True},
            ),
        ]
    )


def make_logs_container(logs_id: str = "log"):
    return html.Div(
        [
            html.Div(
                id=logs_id,
                className="flex-auto flex-col h-24 w-full general-border overflow-y-auto",
            )
        ],
        className="w-full p-2 align-bottom",
    )


def make_label_container(label: str, children: list):
    norm_label = label.lower().replace(" ", "")
    return html.Div(
        [
            html.Label(
                label, className="font-bold mb-2 ml-3", id=f"{norm_label}-label"
            ),
            html.Div(
                children,
                id=f"{norm_label}-container",
                className="general-container min-h-80",
            ),
        ],
        className="w-full",
    )


def make_label_container_callback(app, label: str):
    label = label.lower().replace(" ", "")

    @app.callback(
        Output(f"{label}-container", "className"),
        Input(f"{label}-label", "n_clicks"),
        State(f"{label}-container", "className"),
    )
    def toggle_depth_map(n_clicks, current_class):
        if n_clicks is None:
            raise PreventUpdate()

        if n_clicks % 2 == 0:
            # remove hidden from current class
            return current_class.replace(" hidden", "")
        return current_class + " hidden"


def make_tabs(
    tab_id: str, tab_names: list, tab_contents: list, outer_class_name: str = "w-full"
):
    assert len(tab_names) == len(tab_contents)
    headers = []
    label_class_name = "font-bold mb-2 ml-3"
    for i, tab_name in enumerate(tab_names):
        class_name = label_class_name
        if i == 0:
            class_name += " underline"
        headers.append(
            html.Label(
                tab_name,
                id={"type": f"tab-label-{tab_id}", "index": i},
                className=class_name,
            )
        )

    contents = []
    container_class_name = "general-container min-h-80"
    for i, tab_content in enumerate(tab_contents):
        class_name = container_class_name
        if i > 0:
            class_name += " hidden"
        contents.append(
            html.Div(
                tab_content,
                id={"type": f"tab-content-{tab_id}", "index": i},
                className=class_name,
            )
        )

    return html.Div(
        [
            html.Div(headers, className="w-full flex justify-start"),
            html.Div(contents, className="w-full"),
        ],
        className=outer_class_name,
    )


def make_tabs_callback(app, tab_id: str):
    @app.callback(
        Output({"type": f"tab-label-{tab_id}", "index": ALL}, "n_clicks"),
        Output({"type": f"tab-label-{tab_id}", "index": ALL}, "className"),
        Output({"type": f"tab-content-{tab_id}", "index": ALL}, "className"),
        Input({"type": f"tab-label-{tab_id}", "index": ALL}, "n_clicks"),
        State({"type": f"tab-label-{tab_id}", "index": ALL}, "className"),
        State({"type": f"tab-content-{tab_id}", "index": ALL}, "className"),
        prevent_initial_call=True,
    )
    def toggle_tab_container(n_clicks, label_class, content_class):
        if n_clicks is None:
            raise PreventUpdate()

        clicked_id = n_clicks.index(1)
        assert clicked_id is not None and clicked_id >= 0 and clicked_id < len(n_clicks)

        for i in range(len(n_clicks)):
            if i == clicked_id:
                label_class[i] += " underline"
                content_class[i] = content_class[i].replace(" hidden", "")
            else:
                label_class[i] = label_class[i].replace(" underline", "")
                content_class[i] = content_class[i] + " hidden"

        return [None] * len(n_clicks), label_class, content_class


def make_segmentation_callbacks(app):
    @app.callback(
        Output(C.IMAGE, "src", allow_duplicate=True),
        Output(C.LOGS_DATA, "data", allow_duplicate=True),
        Input(C.SEG_INVERT_MASK, "n_clicks"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State(C.LOGS_DATA, "data"),
        prevent_initial_call=True,
    )
    def invert_mask(n_clicks, filename, logs):
        if n_clicks is None or filename is None:
            raise PreventUpdate()

        state = AppState.from_cache(filename)
        if state.slice_mask is None:
            shape = (state.imgData.size[1], state.imgData.size[0])
            state.slice_mask = np.zeros(shape, dtype=np.uint8)

        state.slice_mask = 255 - state.slice_mask

        image = state.apply_mask(state.imgData, state.slice_mask)
        logs.append("Inverted mask")

        return image, logs

    @app.callback(
        Output(C.IMAGE, "src", allow_duplicate=True),
        Output(C.LOGS_DATA, "data", allow_duplicate=True),
        Input(C.SEG_FEATHER_MASK, "n_clicks"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State(C.LOGS_DATA, "data"),
        prevent_initial_call=True,
    )
    def blur_mask(n_clicks, filename, logs):
        if n_clicks is None or filename is None:
            raise PreventUpdate()

        state = AppState.from_cache(filename)
        if state.slice_mask is None:
            logs.append("No mask to feather")
            return no_update, logs

        # blur the mask
        feather_amount = 10
        state.slice_mask = cv2.blur(state.slice_mask, (feather_amount, feather_amount))

        image = state.apply_mask(state.imgData, state.slice_mask)
        logs.append(f"Feathered mask by {feather_amount} pixels")

        return image, logs

    @app.callback(
        Output(C.SEG_MULTI_COMMIT, "disabled"),
        Output(C.SEG_MULTI_POINT, "disabled"),
        Input(C.DROPDOWN_MODE_SELECTOR, "value"),
        Input(C.STORE_APPSTATE_FILENAME, "data"),
    )
    def toggle_segmentation_buttons(value, filename):
        if value == "segment" and filename is not None:
            return False, False
        return True, True

    @app.callback(
        Output(C.SEG_MULTI_POINT, "className"),
        Output(C.STORE_CLEAR_PREVIEW, "data", allow_duplicate=True),
        Input(C.SEG_MULTI_POINT, "n_clicks"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State(C.SEG_MULTI_POINT, "className"),
        prevent_initial_call=True,
    )
    def toggle_multi_point(n_clicks, filename, class_name):
        if n_clicks is None or filename is None:
            raise PreventUpdate()

        state = AppState.from_cache(filename)

        selected_color = "color-is-selected"
        deselected_color = "color-not-selected"

        # if it's blue, we'll switch to green and turn on multi-point mode
        state.multi_point_mode = True if deselected_color in class_name else False
        state.points_selected = []
        if deselected_color in class_name:
            class_name = class_name.replace(deselected_color, selected_color)
        else:
            class_name = class_name.replace(selected_color, deselected_color)
        return class_name, True

    @app.callback(
        Output(C.SEG_TOGGLE_CHECKERBOARD, "className"),
        Output("gen-slice-output", "children", allow_duplicate=True),
        Output(C.IMAGE, "src", allow_duplicate=True),
        Input(C.SEG_TOGGLE_CHECKERBOARD, "n_clicks"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State(C.SEG_TOGGLE_CHECKERBOARD, "className"),
        prevent_initial_call=True,
    )
    def toggle_checkerboard(n_clicks, filename, class_name):
        if n_clicks is None or filename is None:
            raise PreventUpdate()

        state = AppState.from_cache(filename)

        selected_color = "color-is-selected"
        deselected_color = "color-not-selected"

        if deselected_color in class_name:
            class_name = class_name.replace(deselected_color, selected_color)
        else:
            class_name = class_name.replace(selected_color, deselected_color)

        state.use_checkerboard = selected_color in class_name

        img_data = no_update
        if state.selected_slice is not None:
            assert state.selected_slice >= 0 and state.selected_slice < len(
                state.image_slices
            )
            mode = (
                CompositeMode.CHECKERBOARD
                if state.use_checkerboard
                else CompositeMode.GRAYSCALE
            )
            img_data = state.serve_slice_image_composed(state.selected_slice, mode=mode)

        return class_name, "", img_data


def make_canvas_callbacks(app):
    @app.callback(
        Output(C.STORE_BOUNDING_BOX, "data", allow_duplicate=True),
        Output(C.LOGS_DATA, "data", allow_duplicate=True),
        Input(C.CANVAS_DATA, "data"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State(C.SLIDER_MASK_PADDING, "value"),
        State(C.CHECKLIST_REGION_OF_INTEREST, "value"),
        State(C.LOGS_DATA, "data"),
        prevent_initial_call=True,
    )
    def save_slice_mask(data, filename, padding, crop, logs):
        if data is None or filename is None:
            raise PreventUpdate()

        state = AppState.from_cache(filename)

        if state.selected_slice is None:
            logs.append("No slice selected to save mask")
            return no_update, logs

        if data == "":
            mask_filename = state.mask_filename(state.selected_slice)
            if Path(mask_filename).exists():
                Path(mask_filename).unlink()
                logs.append(f"Deleted mask for slice {state.selected_slice}")
            return no_update, logs

        # turn the data url into a RGBA PIL image
        image = Image.open(io.BytesIO(base64.b64decode(data.split(",")[1])))

        # Split the image into individual channels
        r, g, b, a = image.split()

        # Create a grayscale image with the alpha channel
        new_image = a

        # Scale new image to the same dimensions as imgData
        new_image = new_image.resize(state.imgData.size, resample=Image.BICUBIC)

        mask_filename = state.save_image_mask(state.selected_slice, new_image)

        # communicate the bounding box to the javascript client where we can visualize it
        bounding_box = no_update
        if "crop" in crop:
            bounding_box = find_square_bounding_box(new_image, padding=padding)

        logs.append(f"Saved mask for slice {state.selected_slice} to {mask_filename}")

        return bounding_box, logs

    @app.callback(
        Output(C.CANVAS_MASK_DATA, "data"),
        Output(C.LOGS_DATA, "data", allow_duplicate=True),
        Input(C.BTN_LOAD_CANVAS, "n_clicks"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State(C.LOGS_DATA, "data"),
        prevent_initial_call=True,
    )
    def load_canvas_mask(n_clicks, filename, logs):
        if n_clicks is None or filename is None:
            raise PreventUpdate()

        state = AppState.from_cache(filename)
        if state.selected_slice is None:
            logs.append("No slice selected to load mask")
            return no_update, logs

        index = state.selected_slice
        mask_filename = state.mask_filename(index)
        if not Path(mask_filename).exists():
            print(f"Mask file {mask_filename} does not exist")
            logs.append(f"Mask file {mask_filename} does not exist")
            return no_update, logs

        print(f"Loading mask for slice {state.selected_slice}")
        logs.append(f"Loading mask for slice {state.selected_slice}")
        mask = Image.open(mask_filename).convert("RGB")

        r, _, _ = mask.split()

        width, height = r.size
        zero_channel = Image.new("L", (width, height))
        new_mask = Image.merge("RGBA", (r, zero_channel, zero_channel, r))

        return to_image_url(new_mask), logs

    app.clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="canvas_load"),
        Output(C.STORE_IGNORE, "data", allow_duplicate=True),
        Input(C.CANVAS_MASK_DATA, "data"),
        prevent_initial_call=True,
    )

    app.clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="canvas_draw"),
        Output(C.CANVAS_DATA, "data", allow_duplicate=True),
        Input(C.CANVAS_PAINT, "event"),
        prevent_initial_call=True,
    )

    app.clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="canvas_clear"),
        Output(C.CANVAS_DATA, "data"),
        Input(C.BTN_CLEAR_CANVAS, "n_clicks"),
    )

    app.clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="canvas_clear"),
        Output(C.STORE_IGNORE, "data"),
        # XXX - this will kill the canvas during inpainting - bad
        Input(C.IMAGE, "src"),
    )

    app.clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="canvas_toggle_erase"),
        Output(C.BTN_ERASE_MODE, "className"),
        Input(C.BTN_ERASE_MODE, "n_clicks"),
        prevent_initial_call=True,
    )


def make_navigation_callbacks(app):
    @app.callback(
        Output(C.IMAGE, "src", allow_duplicate=True),
        Output(C.LOGS_DATA, "data", allow_duplicate=True),
        Output(C.STORE_INPAINTING, "data", allow_duplicate=True),
        Input(C.NAV_RESET, "n_clicks"),
        Input(C.NAV_UP, "n_clicks"),
        Input(C.NAV_DOWN, "n_clicks"),
        Input(C.NAV_LEFT, "n_clicks"),
        Input(C.NAV_RIGHT, "n_clicks"),
        Input(C.NAV_ZOOM_IN, "n_clicks"),
        Input(C.NAV_ZOOM_OUT, "n_clicks"),
        State(C.STORE_APPSTATE_FILENAME, "data"),
        State(C.LOGS_DATA, "data"),
        prevent_initial_call=True,
    )
    def navigate_image(reset, up, down, left, right, zoom_in, zoom_out, filename, logs):
        if filename is None:
            raise PreventUpdate()

        nav_clicked = ctx.triggered_id
        if nav_clicked is None:
            raise PreventUpdate()

        state = AppState.from_cache(filename)
        state.selected_slice = None

        if len(state.image_slices) == 0:
            logs.append("No image slices to navigate")
            return no_update, logs, True

        camera_position = state.camera.camera_position

        if nav_clicked == C.NAV_RESET:
            camera_position = np.array(
                [0, 0, -state.camera.camera_distance], dtype=np.float32
            )
        else:
            # Move the camera position based on the navigation button clicked
            # The distance should be configurable
            switch = {
                C.NAV_UP: np.array([0, -1, 0], dtype=np.float32),
                C.NAV_DOWN: np.array([0, 1, 0], dtype=np.float32),
                C.NAV_LEFT: np.array([-1, 0, 0], dtype=np.float32),
                C.NAV_RIGHT: np.array([1, 0, 0], dtype=np.float32),
                C.NAV_ZOOM_OUT: np.array([0, 0, -1], dtype=np.float32),
                C.NAV_ZOOM_IN: np.array([0, 0, 1], dtype=np.float32),
            }

            camera_position += switch[nav_clicked]

        state.camera.camera_position = camera_position

        camera_matrix = state.camera_matrix()
        card_corners_3d_list = state.get_cards()

        image = render_view(
            state.image_slices, camera_matrix, card_corners_3d_list, camera_position
        )

        logs.append(f"Navigated to new camera position {camera_position}")

        return state.serve_main_image(image), logs, True


def make_mode_selector():
    return html.Div(
        [
            html.Label("Mode Selector", className="font-bold mb-2 ml-3"),
            html.Div(
                [
                    html.Div(
                        [
                            dcc.Dropdown(
                                id=C.DROPDOWN_MODE_SELECTOR,
                                options=[
                                    {"label": "Depth Map", "value": "depth"},
                                    {
                                        "label": "Instance Segmentation",
                                        "value": "segment",
                                    },
                                ],
                                value="depth",
                                className="general-dropdown",
                            ),
                        ],
                        className="w-full",
                    ),
                    html.Div(
                        [
                            """
            Switch between depth map and instance segmentation.
            Depth map allows the creation of slices from bands of depth based on the depth map.
            Instance segmentation allows the creation of slices from selected objects on the image.
            """
                        ],
                        className="w-full",
                    ),
                ],
                className="w-full grid grid-cols-2 gap-2 p-2",
            ),
        ],
        className="w-full",
    )



import unittest
from . import constants as C


class TestConstants(unittest.TestCase):

    def test_unique_constants(self):
        constants_dict = {
            name: value for name, value in vars(C).items() if isinstance(value, str)
        }
        values = list(constants_dict.values())
        unique_values = set(values)

        if len(values) != len(unique_values):
            # Find non-unique values
            duplicates = set([value for value in values if values.count(value) > 1])
            non_unique_constants = {
                name: value
                for name, value in constants_dict.items()
                if value in duplicates
            }
            self.fail(f"String constants are not unique: {non_unique_constants}")


if __name__ == "__main__":
    unittest.main()



import numpy as np
import unittest
from .gltf import subdivide_geometry, triangle_indices_from_grid


class TestSubdivideGeometry(unittest.TestCase):
    def test_subdivide_2d(self):
        coords = np.array([[0, 0], [1, 0], [0, 1], [1, 1]], dtype=np.float32)
        subdivisions = 2
        dimension = 2
        expected = np.array(
            [
                [0.0, 0.0],
                [0.5, 0.0],
                [1.0, 0.0],
                [0.0, 0.5],
                [0.5, 0.5],
                [1.0, 0.5],
                [0.0, 1.0],
                [0.5, 1.0],
                [1.0, 1.0],
            ],
            dtype=np.float32,
        )
        result = subdivide_geometry(coords, subdivisions, dimension)
        np.testing.assert_array_equal(result, expected)

    def test_subdivide_3d(self):
        coords = np.array(
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0]], dtype=np.float32
        )
        subdivisions = 2
        dimension = 3
        expected = np.array(
            [
                [0.0, 0.0, 0.0],
                [0.5, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 0.5, 0.0],
                [0.5, 0.5, 0.0],
                [1.0, 0.5, 0.0],
                [0.0, 1.0, 0.0],
                [0.5, 1.0, 0.0],
                [1.0, 1.0, 0.0],
            ],
            dtype=np.float32,
        )
        result = subdivide_geometry(coords, subdivisions, dimension)
        np.testing.assert_array_equal(result, expected)


class TestTriangleIndicesFromGrid(unittest.TestCase):
    def test_triangle_indices_from_grid(self):
        vertices = np.array(
            [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0]], dtype=np.float32
        )
        expected = np.array([[0, 1, 2], [2, 1, 3]], dtype=np.uint32)
        result = triangle_indices_from_grid(vertices)
        np.testing.assert_array_equal(result, expected)


if __name__ == "__main__":
    unittest.main()



import cv2
from pathlib import Path
from PIL import Image
import numpy as np

from .utils import filename_add_version, filename_previous_version
from .camera import Camera


class ImageSlice:
    __slots__ = (
        "_image",
        "_depth",
        "_filename",
        "_is_ground_plane",
        "positive_prompt",
        "negative_prompt",
    )

    def __init__(
        self,
        image=None,
        depth=-1,
        filename=None,
        positive_prompt="",
        negative_prompt="",
    ):
        self.image = image
        self._depth = depth
        self._filename = filename
        self._is_ground_plane = False

        self.positive_prompt = positive_prompt
        self.negative_prompt = negative_prompt

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, value):
        if (
            not isinstance(value, np.ndarray)
            and not isinstance(value, Image.Image)
            and value is not None
        ):
            raise ValueError("image must be a np.ndarray, Image.Image object or None")
        if isinstance(value, Image.Image):
            print(
                "WARNING: ImageSlice.image is being set with a PIL Image object. This is not recommended."
            )
            value = np.array(value)
        self._image = value

    @property
    def is_ground_plane(self):
        return self._is_ground_plane

    @is_ground_plane.setter
    def is_ground_plane(self, value):
        if not isinstance(value, bool):
            raise ValueError("is_ground_plane must be a boolean value")
        self._is_ground_plane = value

    @property
    def depth(self):
        return self._depth

    @depth.setter
    def depth(self, value):
        if not isinstance(value, (float, int)) or value < -1:
            raise ValueError(
                "depth must be a non-negative number or -1 for unknown depth"
            )
        self._depth = value

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        if (
            not isinstance(value, Path)
            and not isinstance(value, str)
            and value is not None
        ):
            raise ValueError("filename must be a Path, str object or None")
        self._filename = value

    def __eq__(self, other):
        if not isinstance(other, ImageSlice):
            return False
        return (
            np.array_equal(self.image, other.image)
            and self.depth == other.depth
            and self.filename == other.filename
        )

    @staticmethod
    def _dimension_at_depth(z: float, image_height: int, image_width: int, cam: Camera):
        fl_px = cam.focal_length_px(image_width)
        card_width = (image_width * (z + cam.camera_distance)) / fl_px
        card_height = (image_height * (z + cam.camera_distance)) / fl_px
        return card_width, card_height

    def _depth_to_z(self, depth: float, cam: Camera):
        return cam.max_distance * ((255 - depth) / 255.0)

    def create_card(self, image_height: int, image_width: int, cam: Camera):
        if self.is_ground_plane:
            near_z = 0
            far_z = cam.max_distance

            near_width, near_height = self._dimension_at_depth(
                near_z, image_height, image_width, cam
            )
            far_width, far_height = self._dimension_at_depth(
                far_z, image_height, image_width, cam
            )

            card_corners_3d = np.array(
                [
                    [-far_width / 2, -far_height / 2, far_z],
                    [far_width / 2, -far_height / 2, far_z],
                    [near_width / 2, near_height / 2, near_z + 2 * far_z],  # NO IDEA
                    [-near_width / 2, near_height / 2, near_z + 2 * far_z],
                ],
                dtype=np.float32,
            )
        else:
            z = self._depth_to_z(self.depth, cam)

            # Calculate the 3D points of the card corners
            card_width, card_height = self._dimension_at_depth(
                z, image_height, image_width, cam
            )

            card_corners_3d = np.array(
                [
                    [-card_width / 2, -card_height / 2, z],
                    [card_width / 2, -card_height / 2, z],
                    [card_width / 2, card_height / 2, z],
                    [-card_width / 2, card_height / 2, z],
                ],
                dtype=np.float32,
            )

        return card_corners_3d

    def save_image(self):
        slice_image = self.image
        if not isinstance(slice_image, Image.Image):
            slice_image = Image.fromarray(slice_image, mode="RGBA")
        output_image_path = self.filename
        print(f"Saving image slice: {output_image_path}")
        slice_image.save(str(output_image_path))

    def read_image(self):
        img = cv2.imread(str(self.filename), cv2.IMREAD_UNCHANGED)
        self.image = cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)

    def new_version(self, image=None, save=True):
        """
        Creates a new version of the image and saves it if specified.

        Args:
            image (np.ndarray, optional): The new image to be set. Defaults to None.
            save (bool, optional): Whether to save the new image. Defaults to True.

        Returns:
            str: The filename of the new version of the image.
        """
        if image is not None:
            assert isinstance(image, np.ndarray)
            self.image = image
        self.filename = filename_add_version(self.filename)
        if save:
            self.save_image()
        return self.filename

    def can_undo(self, forward=False):
        """
        Check if it is possible to undo the specified image slice.

        Args:
            forward (bool, optional): If True, check for the next version of the slice.
                                      If False, check for the previous version. Defaults to False.

        Returns:
            bool: True if the specified slice version exists, False otherwise.
        """
        if forward:
            filename = filename_add_version(self.filename)
        else:
            filename = filename_previous_version(self.filename)

        if filename is None:
            return False

        return Path(filename).exists()

    def undo(self, forward=False):
        """
        Undo the specified image slice.

        Args:
            forward (bool, optional): If True, undo the next version of the slice.
                          If False, undo the previous version. Defaults to False.

        Returns:
            bool: True if the undo operation is successful, False otherwise.
        """
        if not self.can_undo(forward):
            return False

        if forward:
            filename = filename_add_version(self.filename)
        else:
            filename = filename_previous_version(self.filename)

        if filename is None:
            return False

        self.filename = filename
        self.read_image()
        return True



from PIL import Image
from io import BytesIO
import requests
import argparse


class StabilityAI:
    """
    Implements the Stability AI API for generating and inpainting images.

    This API has several limitations:
    - Inpainting masks cannot be feathered or have soft edges. Consequently, the inpainting does not work with low strength.
    """

    MAX_PIXELS = 9437184

    def __init__(self, api_key):
        self.api_key = api_key

    def validate_key(self):
        response = requests.get(
            "https://api.stability.ai/v1/user/balance",
            headers={"Authorization": f"Bearer {self.api_key}"},
        )

        if response.status_code != 200:
            return False, None

        # Do something with the payload...
        payload = response.json()
        assert "credits" in payload, f"Invalid API response: {payload}."
        return True, payload["credits"]

    def generate_image(
        self, prompt, negative_prompt="", aspect_ratio="16:9", output_format="png"
    ):
        """
        Generates an image using the Stability AI API.

        Args:
            prompt (str): The main text prompt for generating the image.
            negative_prompt (str, optional): The negative text prompt for generating the image. Defaults to ''.
            aspect_ratio (str, optional): The aspect ratio of the generated image. Defaults to "16:9".
            output_format (str, optional): The output format of the generated image. Defaults to "png".

        Returns:
            PIL.Image.Image: The generated image as a PIL Image object.

        Raises:
            Exception: If the API response status code is not 200.
        """
        response = requests.post(
            f"https://api.stability.ai/v2beta/stable-image/generate/core",
            headers={"authorization": f"Bearer {self.api_key}", "accept": "image/*"},
            files={"none": ""},
            data={
                "mode": "text-to-image",
                "prompt": prompt,
                "negative_prompt": negative_prompt,
                "aspect_ratio": aspect_ratio,
                "output_format": output_format,
            },
        )

        if response.status_code != 200:
            raise Exception(str(response.json()))

        return Image.open(BytesIO(response.content))

    def image_to_image(
        self, image, prompt, negative_prompt="", strength=0.8, output_format="png"
    ):
        """
        Generates an image-to-image request using the Stability AI API.

        Args:
            image (PIL.Image.Image): The input image to be transformed.
            prompt (str): The prompt to guide the inpainting process.
            negative_prompt (str, optional): The negative prompt to guide the inpainting process. Defaults to ''.
            strength (float, optional): The strength of the transformation effect. Defaults to 0.8.
            output_format (str, optional): The format of the output image. Defaults to "png".

        Returns:
            PIL.Image.Image: The inpainted image.

        Raises:
            ValueError: If the image is smaller than 64x64 pixels.
            Exception: If the API request fails.

        """
        # check that the image it at least 64x64
        if image.width < 64 or image.height < 64:
            raise ValueError("Image must be at least 64x64")

        orig_width, orig_height = image.width, image.height

        image = self._resize_image(image)

        # prepare image and mask data
        image_data = BytesIO()
        image.save(image_data, format="PNG")
        image_data = image_data.getvalue()

        response = requests.post(
            f"https://api.stability.ai/v2beta/stable-image/generate/sd3",
            headers={"authorization": f"Bearer {self.api_key}", "accept": "image/*"},
            files={"image": image_data},
            data={
                "prompt": prompt,
                "negative_prompt": negative_prompt,
                "mode": "image-to-image",
                "strength": strength,
                "output_format": output_format,
            },
        )

        if response.status_code != 200:
            raise Exception(str(response.json()))

        image = Image.open(BytesIO(response.content))
        if orig_width != image.width or orig_height != image.height:
            image = image.resize((orig_width, orig_height), Image.LANCZOS)
        return image

    def _resize_image(self, image, max_pixels=MAX_PIXELS):
        if image.width * image.height > max_pixels:
            # square root as with multiply the ratio twice below
            ratio = (max_pixels / (image.width * image.height)) ** 0.5
            new_width = int(image.width * ratio)
            new_height = int(image.height * ratio)
            image = image.resize((new_width, new_height), Image.LANCZOS)
        return image

    def inpaint_image(
        self, image, mask, prompt, negative_prompt="", strength=1.0, output_format="png"
    ):
        """
        Inpaints an image using the Stability AI API.

        Args:
            image (PIL.Image.Image): The input image to be inpainted.
            mask (PIL.Image.Image): The mask indicating the areas to be inpainted.
            prompt (str): The prompt to guide the inpainting process.
            negative_prompt (str, optional): The negative prompt to guide the inpainting process. Defaults to ''.
            strength (float, optional): The strength of the inpainting effect. Defaults to 1.0.
            output_format (str, optional): The format of the output image. Defaults to "png".

        Returns:
            PIL.Image.Image: The inpainted image.

        Raises:
            ValueError: If the image is smaller than 64x64 pixels.
            Exception: If the API request fails.

        """
        # check that the image it at least 64x64
        if image.width < 64 or image.height < 64:
            raise ValueError("Image must be at least 64x64")

        orig_width, orig_height = image.width, image.height

        image = self._resize_image(image)
        mask = self._resize_image(mask)

        # scale the mask by strength - the API does not work well with low strength
        if strength < 1.0:
            mask = mask.point(lambda p: p * strength)

        # prepare image and mask data
        image_data = BytesIO()
        image.save(image_data, format="PNG")
        image_data = image_data.getvalue()
        mask_data = BytesIO()
        mask.save(mask_data, format="PNG")
        mask_data = mask_data.getvalue()

        response = requests.post(
            f"https://api.stability.ai/v2beta/stable-image/edit/inpaint",
            headers={"authorization": f"Bearer {self.api_key}", "accept": "image/*"},
            files={"image": image_data, "mask": mask_data},
            data={
                "prompt": prompt,
                "negative_prompt": negative_prompt,
                "output_format": output_format,
            },
        )

        if response.status_code != 200:
            raise Exception(str(response.json()))

        image = Image.open(BytesIO(response.content))
        if orig_width != image.width or orig_height != image.height:
            image = image.resize((orig_width, orig_height), Image.LANCZOS)
        return image

    def upscale_image(self, image, prompt, negative_prompt="", output_format="png"):
        # check that the image it at least 64x64
        if image.width < 64 or image.height < 64:
            raise ValueError("Image must be at least 64x64")

        if image.width * image.height > 1024 * 1024:
            raise ValueError(
                f"Image must be at most 1024x1024: Got {image.width}x{image.height} = {image.width * image.height} pixels."
            )

        # prepare image and mask data
        image_data = BytesIO()
        image.save(image_data, format="PNG")
        image_data = image_data.getvalue()

        response = requests.post(
            f"https://api.stability.ai/v2beta/stable-image/upscale/conservative",
            headers={"authorization": f"Bearer {self.api_key}", "accept": "image/*"},
            files={"image": image_data},
            data={
                "prompt": prompt,
                "negative_prompt": negative_prompt,
                "output_format": output_format,
            },
        )

        if response.status_code != 200:
            raise Exception(str(response.json()))

        upscaled_image = Image.open(BytesIO(response.content))
        return upscaled_image


def main():  # pragma: no cover
    parser = argparse.ArgumentParser()
    parser.add_argument("--api-key", required=True)
    parser.add_argument(
        "-p",
        "--prompt",
        default="an intelligent AI robot on the surface of the moon; intelligent and cute glowing eyes; "
        + "surrounded by canyons and rocks; dystopian atmosphere; scfi vibe; bladerunner and cyberpunk; "
        + "distant stars visible in the sky behind the mountains; photorealistic; everything in sharp focus",
    )
    parser.add_argument("--aspect-ratio", default="16:9")
    parser.add_argument("--output-format", default="png")
    parser.add_argument("-i", "--image", type=str, default=None)
    parser.add_argument("-m", "--mask", type=str, default=None)
    parser.add_argument("-s", "--strength", type=float, default=0.8)
    parser.add_argument("-u", "--upscale", action="store_true")
    args = parser.parse_args()

    ai = StabilityAI(args.api_key)
    success, _ = ai.validate_key()
    if not success:
        print("Invalid API key.")
        return
    if args.image and args.upscale:
        image = Image.open(args.image)
        image = ai._resize_image(image, 1024 * 1024)
        image = ai.upscale_image(image, args.prompt)
    elif args.image and args.mask:
        image = Image.open(args.image)
        mask = Image.open(args.mask)

        image = ai.inpaint_image(image, mask, args.prompt, strength=1.0)
    elif args.image:
        print("Running image to image")
        image = Image.open(args.image)
        image = ai.image_to_image(image, args.prompt, strength=args.strength)
    else:
        image = ai.generate_image(
            args.prompt,
            aspect_ratio=args.aspect_ratio,
            output_format=args.output_format,
        )
    image.save("output.png")


if __name__ == "__main__":
    main()



# (c) 2024 Niels Provos
#
"""
Create Depth Maps from Images

This module provides functionality to create depth maps from images using pre-trained deep learning models.
The depth maps can be used to create parallax effects in images and videos.

TODO:
 - Investigate https://huggingface.co/LiheYoung/depth_anything_vitl14 - also at https://huggingface.co/docs/transformers/main/model_doc/depth_anything
"""


from transformers import AutoImageProcessor, DPTForDepthEstimation
import torch
import cv2
import numpy as np
from PIL import Image

from .utils import torch_get_device


class DepthEstimationModel:
    MODELS = ["midas", "zoedepth", "dinov2"]

    def __init__(self, model="midas"):
        assert model in self.MODELS, f"Model {model} must be one of {self.MODELS}"
        self._model_name = model
        self.model = None
        self.transforms = None
        self.image_processor = None

    def __eq__(self, other):
        if not isinstance(other, DepthEstimationModel):
            return False
        return self._model_name == other._model_name

    @property
    def model_name(self):
        return self._model_name

    def load_model(self, progress_callback=None):
        load_pipeline = {
            "midas": create_medias_pipeline,
            "zoedepth": create_zoedepth_pipeline,
            "dinov2": create_dinov2_pipeline,
        }

        result = load_pipeline[self._model_name](progress_callback=progress_callback)

        if self._model_name == "midas":
            self.model, self.transforms = result
        elif self._model_name == "zoedepth":
            self.model = result
        elif self._model_name == "dinov2":
            self.model, self.image_processor = result

    def depth_map(self, image, progress_callback=None):
        if self.model is None:
            self.load_model()

        run_pipeline = {
            "midas": lambda img, cb: run_medias_pipeline(
                img, self.model, self.transforms, progress_callback=cb
            ),
            "zoedepth": lambda img, cb: run_zoedepth_pipeline(
                img, self.model, progress_callback=cb
            ),
            "dinov2": lambda img, cb: run_dinov2_pipeline(
                img, self.model, self.image_processor, progress_callback=cb
            ),
        }

        return run_pipeline[self._model_name](image, progress_callback)


def create_dinov2_pipeline(progress_callback=None):
    image_processor = AutoImageProcessor.from_pretrained(
        "facebook/dpt-dinov2-large-nyu"
    )
    model = DPTForDepthEstimation.from_pretrained("facebook/dpt-dinov2-large-nyu")
    model.to(torch_get_device())
    return model, image_processor


def run_dinov2_pipeline(image, model, image_processor, progress_callback=None):
    image = Image.fromarray(image)

    new_size = image.size
    if image.width > image.height:
        if image.width > 1024:
            new_size = (1024, int(image.height * 1024 / image.width))
    else:
        if image.height > 1024:
            new_size = (int(image.width * 1024 / image.height), 1024)

    resized_image = image.convert("RGB").resize(new_size, Image.BICUBIC)
    inputs = image_processor(images=resized_image, return_tensors="pt")
    inputs = {k: v.to(torch_get_device()) for k, v in inputs.items()}
    with torch.no_grad():
        outputs = model(**inputs)
        predicted_depth = outputs.predicted_depth

    # interpolate to original size
    prediction = torch.nn.functional.interpolate(
        predicted_depth.unsqueeze(1),
        size=image.size[::-1],
        mode="bicubic",
        align_corners=False,
    )

    # visualize the prediction
    output = prediction.squeeze().cpu().numpy()
    formatted = (output * 255 / np.max(output)).astype("uint8")
    formatted[:, :] = 255 - formatted[:, :]  # invert the depth map

    # resize to original size
    formatted = cv2.resize(
        formatted, (image.width, image.height), interpolation=cv2.INTER_CUBIC
    )

    return formatted


def create_medias_pipeline(progress_callback=None):
    """
    Creates a media pipeline using the MiDaS model for depth estimation.

    Args:
        progress_callback (callable, optional): A callback function to report progress. Defaults to None.

    Returns:
        tuple: A tuple containing the MiDaS model and the transformation pipeline.

    """
    # Load the MiDaS v2.1 model
    model_type = "DPT_Large"
    midas = torch.hub.load("intel-isl/MiDaS", model_type, skip_validation=True)

    if progress_callback:
        progress_callback(30, 100)

    # Set the model to evaluation mode
    midas.eval()

    # Define the transformation pipeline
    midas_transforms = torch.hub.load(
        "intel-isl/MiDaS", "transforms", skip_validation=True
    )
    if model_type == "DPT_Large" or model_type == "DPT_Hybrid":
        transforms = midas_transforms.dpt_transform
    else:
        transforms = midas_transforms.small_transform

    if progress_callback:
        progress_callback(50, 100)

    # Set the device (CPU or GPU)
    midas.to(torch_get_device())

    return midas, transforms


def run_medias_pipeline(image, midas, transforms, progress_callback=None):
    """
    Runs the media pipeline for segmentation.

    Args:
        image (numpy.ndarray): The input image.
        midas (torch.nn.Module): The MIDAS model.
        transforms (torchvision.transforms.Compose): The image transforms.
        progress_callback (callable, optional): A callback function to report progress.

    Returns:
        numpy.ndarray: The predicted segmentation mask.
    """
    input_batch = transforms(image).to(torch_get_device())
    with torch.no_grad():
        prediction = midas(input_batch)

        prediction = torch.nn.functional.interpolate(
            prediction.unsqueeze(1),
            size=image.shape[:2],
            mode="bicubic",
            align_corners=False,
        ).squeeze()

    if progress_callback:
        progress_callback(90, 100)

    return prediction.cpu().numpy()


def midas_depth_map(image, progress_callback=None):
    if progress_callback:
        progress_callback(0, 100)

    midas, transforms = create_medias_pipeline(progress_callback=progress_callback)

    depth_map = run_medias_pipeline(
        image, midas, transforms, progress_callback=progress_callback
    )

    if progress_callback:
        progress_callback(100, 100)

    return depth_map


def create_zoedepth_pipeline(progress_callback=None):
    # Triggers fresh download of MiDaS repo
    torch.hub.help("intel-isl/MiDaS", "DPT_BEiT_L_384", force_reload=True)

    # Zoe_NK
    model_zoe_nk = torch.hub.load(
        "isl-org/ZoeDepth", "ZoeD_NK", pretrained=True, skip_validation=True
    )

    # Set the device (CPU or GPU)
    device = torch_get_device()
    model_zoe_nk.to(device)

    if progress_callback:
        progress_callback(50, 100)

    return model_zoe_nk


def run_zoedepth_pipeline(image, model_zoe_nk, progress_callback=None):
    depth_map = model_zoe_nk.infer_pil(image)  # as numpy

    # invert the depth map since we are expecting the farthest objects to be black
    depth_map = 255 - depth_map

    if progress_callback:
        progress_callback(100, 100)

    return depth_map


def zoedepth_depth_map(image, progress_callback=None):
    model_zoe_nk = create_zoedepth_pipeline(progress_callback=progress_callback)

    return run_zoedepth_pipeline(
        image, model_zoe_nk, progress_callback=progress_callback
    )



import base64
import requests
from PIL import Image
from io import BytesIO
import json
import argparse

from .utils import to_image_url

MAX_DIMENSION = 2048


def create_img2img_payload(
    input_image,
    positive_prompt,
    negative_prompt,
    mask_image=None,
    strength=0.8,
    steps=20,
    cfg_scale=8.0,
):
    """
    Create a payload for the img2img API.

    Args:
        input_image (PIL.Image.Image): The input image.
        positive_prompt (str): The positive prompt for generating the output image.
        negative_prompt (str): The negative prompt for generating the output image.
        mask_image (PIL.Image.Image, optional): The mask for the input image. Defaults to None.
        strength (float, optional): The denoising strength. Defaults to 0.8.
        steps (int, optional): The number of steps for generating the output image. Defaults to 20.
        cfg_scale (float, optional): The scale factor for the configuration. Defaults to 8.0.

    Returns:
        dict: The payload for the img2img API.
    """
    width, height = input_image.size
    if width > MAX_DIMENSION or height > MAX_DIMENSION:
        # resize the image so the largest dimension is MAX_DIMENSION
        if width > height:
            new_width = MAX_DIMENSION
            new_height = int(MAX_DIMENSION * height / width)
        else:
            new_height = MAX_DIMENSION
            new_width = int(MAX_DIMENSION * width / height)
        input_image = input_image.resize((new_width, new_height))
        mask_image = mask_image.resize((new_width, new_height))

    payload = {
        "init_images": [to_image_url(input_image)],
        "prompt": positive_prompt,
        "negative_prompt": negative_prompt,
        "denoising_strength": strength,
        "width": width,
        "height": height,
        "steps": steps,
        "cfg_scale": cfg_scale,
        "mask_blur_x": 0,  # if we provide a mask it's already blurred
        "mask_blur_y": 0,
        "mask_blur": 0,
    }

    if mask_image is not None:
        payload["mask"] = to_image_url(mask_image)

    return payload


def make_img2img_request(server_address, payload):
    """
    Sends an img2img request to the automatic1111 API endpoint.

    Args:
        server_address (str): The address:port of the server to send the request to.
        payload (dict): The payload generated by create_img2img_payload.

    Returns:
        list: A list of images returned in the response.

    Raises:
        requests.exceptions.Timeout: If the request times out.

    """
    server_url = f"http://{server_address}/sdapi/v1/img2img"
    payload = json.dumps(payload)
    response = requests.post(url=server_url, data=payload, timeout=500).json()

    return response.get("images")


def make_models_request(server_address):
    """
    Sends a request to the automatic1111 API endpoint to get the available models.

    Args:
        server_address (str): The address:port of the server to send the request to.

    Returns:
        list: A list of available models.

    Raises:
        requests.exceptions.Timeout: If the request times out.

    """
    server_url = f"http://{server_address}/sdapi/v1/sd-models"
    response = requests.get(url=server_url, timeout=3).json()

    if not isinstance(response, list):
        return None

    return [entry["model_name"] for entry in response]


def main():
    # Parse arguments - input image, positive prompt, negative prompt, server address
    parser = argparse.ArgumentParser(description="Process some integers.")
    parser.add_argument("-i", "--input-image", type=str, help="Path to the input image")
    parser.add_argument(
        "-m", "--mask-image", type=str, default=None, help="Path to the mask image"
    )
    parser.add_argument("-p", "--prompt", type=str, help="Positive prompt")
    parser.add_argument(
        "-n", "--negative-prompt", type=str, default="", help="Negative prompt"
    )
    parser.add_argument("-e", "--steps", type=int, default=30, help="Number of steps")
    parser.add_argument("-s", "--server", type=str, help="Server address")
    args = parser.parse_args()

    image = Image.open(args.input_image)
    mask_image = None
    if args.mask_image is not None:
        mask_image = Image.open(args.mask_image)

    payload = create_img2img_payload(
        image,
        args.prompt,
        args.negative_prompt,
        mask_image=mask_image,
        steps=args.steps,
    )
    images = make_img2img_request(args.server, payload)
    if images is None:
        print("No response from server")
    else:
        for image in images:
            image = Image.open(BytesIO(base64.b64decode(image)))
            image.show()
            input("Press Enter to continue...")


if __name__ == "__main__":
    main()



import unittest
from unittest.mock import patch, MagicMock
from dash import dcc, html, no_update
from dash.exceptions import PreventUpdate
from PIL import Image
import numpy as np
from pathlib import Path

from .webui import (
    update_threshold_values,
    click_event,
    copy_to_clipboard,
    export_state_as_gltf,
    slice_upload,
    update_slices,
)
from .controller import AppState
from .utils import to_image_url
from .camera import Camera
from .slice import ImageSlice
from . import constants as C


class TestUpdateThresholds(unittest.TestCase):
    def test_update_threshold_values_boundaries(self):
        # Create a dummy state
        state = AppState()
        state.imgThresholds = [0, 10, 20, 30, 40, 255]

        # set up a fake cache
        filename = "teststate"
        state.cache[filename] = state

        # Call the function
        input_thresholds = [0, 20, 30, 40, 255]
        num_slices = 5
        threshold_values, _ = update_threshold_values(
            input_thresholds, num_slices, filename
        )

        # Assert that the state is updated
        self.assertEqual(state.imgThresholds, [0, 1, 20, 30, 40, 254, 255])

    def test_update_threshold_values_limit(self):
        # Create a dummy state
        state = AppState()
        state.imgThresholds = [0, 10, 20, 30, 40, 255]

        # set up a fake cache
        filename = "teststate"
        state.cache[filename] = state

        # Call the function
        input_thresholds = [255, 255, 255, 255, 255]
        num_slices = 5
        threshold_values, _ = update_threshold_values(
            input_thresholds, num_slices, filename
        )

        # Assert that the state is updated
        self.assertEqual(state.imgThresholds, [0, 255, 256, 257, 258, 254, 255])


class TestClickEvent(unittest.TestCase):

    def setUp(self):
        # Patch objects and methods that aren't the focus of this test
        self.ctx_patch = patch("parallax_maker.webui.ctx")
        self.AppState_patch = patch("parallax_maker.webui.AppState")
        self.find_pixel_patch = patch("parallax_maker.webui.find_pixel_from_event")
        self.SegmentationModel_patch = patch("parallax_maker.webui.SegmentationModel")
        self.no_update_patch = patch("parallax_maker.webui.no_update")

        self.mock_ctx = self.ctx_patch.start()
        self.mock_AppState = self.AppState_patch.start()
        self.mock_find_pixel = self.find_pixel_patch.start()
        self.mock_SegmentationModel = self.SegmentationModel_patch.start()
        self.mock_no_update = self.no_update_patch.start()

        self.mock_state = self.mock_AppState.from_cache.return_value
        self.mock_segmentation_model = self.mock_SegmentationModel.return_value

        # Define default mock return values
        self.mock_state.depth_slice_from_pixel.return_value = (np.ones((100, 100)), 1)
        self.mock_segmentation_model.mask_at_point_blended.return_value = np.ones(
            (100, 100)
        )

        self.mock_image = Image.new("RGB", (100, 100))
        self.mock_mask = np.ones((100, 100))

    def tearDown(self):
        # Stop patches
        patch.stopall()

    def test_click_event_no_filename(self):
        with self.assertRaises(PreventUpdate):
            click_event(None, None, None, None, None, None, None)

    @patch("builtins.print")
    def test_click_event_invalid_trigger(self, mock_print):
        self.mock_ctx.triggered_id = "invalid_trigger"
        with self.assertRaises(ValueError):
            click_event(None, None, None, None, None, "filename", [])

    def test_click_event_no_element_or_data(self):
        self.mock_ctx.triggered_id = "el"
        with self.assertRaises(PreventUpdate):
            click_event(None, None, None, None, None, "filename", [])

    def test_click_event_segment_mode_shift_click(self):
        self.mock_ctx.triggered_id = "el"
        self.mock_state.multi_point_mode = False
        self.mock_state.imgData = self.mock_image
        self.mock_state.slice_mask = self.mock_mask
        self.mock_state.segmentation_model = None
        element = {"shiftKey": True, "ctrlKey": False}
        rect_data = "rect_data"

        self.mock_find_pixel.return_value = (10, 10)

        result = click_event(None, None, element, rect_data, "segment", "filename", [])

        self.mock_state.apply_mask.assert_called_once()
        self.assertEqual(result[0], self.mock_state.serve_main_image.return_value)

    def test_click_event_no_shift_or_ctrl_click(self):
        self.mock_ctx.triggered_id = C.SEG_MULTI_COMMIT
        self.mock_state.multi_point_mode = True
        self.mock_state.points_selected = [((10, 10), False)]
        self.mock_state.imgData = "image_data"
        self.mock_state.segmentation_model = None

        element = {"shiftKey": False, "ctrlKey": False}
        rect_data = "rect_data"

        result = click_event(None, None, element, rect_data, "segment", "filename", [])

        self.assertEqual(
            result[1], ["Committed points [(10, 10)] and [] for Segment Anything"]
        )

    def test_click_event_apply_mask(self):
        self.mock_ctx.triggered_id = "el"
        self.mock_state.multi_point_mode = False
        self.mock_state.imgData = "image_data"
        self.mock_state.slice_mask = self.mock_mask
        element = {"shiftKey": True, "ctrlKey": False}
        rect_data = "rect_data"

        self.mock_find_pixel.return_value = (10, 10)

        result = click_event(None, None, element, rect_data, "mode", "filename", [])

        self.mock_state.apply_mask.assert_called_once()
        self.assertEqual(result[0], self.mock_state.serve_main_image.return_value)

    def test_click_event_with_slice(self):
        self.mock_ctx.triggered_id = "el"
        self.mock_state.multi_point_mode = False
        self.mock_state.imgData = "image_data"
        self.mock_state.slice_mask = self.mock_mask
        self.mock_state.image_slices = [np.ones((100, 100, 4))]
        self.mock_state.selected_slice = 0
        element = {"shiftKey": True, "ctrlKey": False}
        rect_data = "rect_data"

        self.mock_find_pixel.return_value = (10, 10)

        result = click_event(None, None, element, rect_data, "mode", "filename", [])

        self.mock_state.apply_mask.assert_called_once()
        self.assertEqual(result[0], self.mock_state.serve_main_image.return_value)


class TestCopyToClipboard(unittest.TestCase):

    @patch("parallax_maker.webui.AppState.from_cache")
    def test_copy_to_clipboard_no_clicks(self, mock_from_cache):
        # Test when n_clicks is None, should raise PreventUpdate
        with self.assertRaises(PreventUpdate):
            copy_to_clipboard(None, "some_filename", [])

    @patch("parallax_maker.webui.AppState.from_cache")
    def test_copy_to_clipboard_no_filename(self, mock_from_cache):
        # Test when filename is None, should raise PreventUpdate
        with self.assertRaises(PreventUpdate):
            copy_to_clipboard(1, None, [])

    @patch("parallax_maker.webui.AppState.from_cache")
    def test_copy_to_clipboard_no_mask_selected(self, mock_from_cache):
        # Mock AppState with no slice_mask
        mock_state = MagicMock()
        mock_state.slice_mask = None
        mock_from_cache.return_value = mock_state

        # Test when no mask is selected
        logs = []
        result = copy_to_clipboard(1, "some_filename", logs)
        self.assertEqual(result, ["No mask selected"])

    @patch("parallax_maker.webui.AppState.from_cache")
    def test_copy_to_clipboard_with_mask_and_slice(self, mock_from_cache):
        # Mock AppState with a slice_mask and a selected slice
        mock_state = MagicMock()
        mock_state.slice_mask = np.zeros((100, 100))
        mock_state.selected_slice = "selected_slice"
        mock_state.slice_image_composed.return_value = MagicMock(
            convert=lambda mode: Image.new("RGBA", (100, 100))
        )

        mock_from_cache.return_value = mock_state

        logs = []
        result = copy_to_clipboard(1, "some_filename", logs)
        self.assertEqual(result, ["Copied mask to clipboard"])
        self.assertTrue(mock_state.clipboard_image is not None)
        np.testing.assert_array_equal(
            mock_state.clipboard_image[:, :, 3], mock_state.slice_mask
        )

    @patch("parallax_maker.webui.AppState.from_cache")
    def test_copy_to_clipboard_with_mask_no_slice(self, mock_from_cache):
        mock_image = Image.new("RGBA", (100, 100))

        # Mock AppState with a slice_mask and no selected slice
        mock_state = MagicMock()
        mock_state.slice_mask = np.zeros((100, 100))
        mock_state.selected_slice = None
        mock_state.imgData = mock_image

        mock_from_cache.return_value = mock_state

        logs = []
        result = copy_to_clipboard(1, "some_filename", logs)
        self.assertEqual(result, ["Copied mask to clipboard"])
        self.assertTrue(mock_state.clipboard_image is not None)
        np.testing.assert_array_equal(
            mock_state.clipboard_image[:, :, 3], mock_state.slice_mask
        )


class TestExportGltf(unittest.TestCase):
    def setUp(self):
        self.state = MagicMock()
        self.state.image_slices = [
            ImageSlice(np.zeros((100, 100, 4), dtype=np.uint8), i) for i in range(3)
        ]

        # mocking depthmap file requires both exists and return_value
        self.mock_depth_file = MagicMock()
        self.state.depth_filename.return_value = self.mock_depth_file
        self.mock_depth_file.exists.return_value = True

        self.camera = Camera(10, 100, 50)

        self.state.camera_matrix.return_value = self.camera.camera_matrix(100, 100)
        self.state.get_cards.return_value = [
            slice.create_card(100, 100, self.camera)
            for slice in self.state.image_slices
        ]

        self.state.upscaled_filename.return_value = Path("upscaled_file.png")
        self.state.MODEL_FILE = "model.gltf"
        self.state.camera = self.camera

    @patch("parallax_maker.webui.generate_depth_map")
    @patch("parallax_maker.webui.postprocess_depth_map")
    @patch("parallax_maker.webui.export_gltf")
    def test_export_state_as_gltf(
        self, mock_export_gltf, mock_postprocess_depth_map, mock_generate_depth_map
    ):
        # Test case 1: Displacement scale is 0
        state = AppState()
        state.image_slices = self.state.image_slices

        mock_export_gltf.return_value = Path("output.gltf")

        result = export_state_as_gltf(
            self.state, "output_dir", Camera(10, 100, 50), 0, "midas"
        )

        self.assertEqual(result, Path("output.gltf"))
        mock_generate_depth_map.assert_not_called()
        mock_postprocess_depth_map.assert_not_called()

        # Compare individual elements of card_corners_3d_list
        expected_call = mock_export_gltf.call_args_list[0]
        expected_args, expected_kwargs = expected_call
        self.assertEqual(expected_args[0], Path("output_dir/model.gltf"))
        image_slices_filenames = [slice.filename for slice in self.state.image_slices]
        self.assertEqual(expected_args[3], image_slices_filenames)
        self.assertEqual(expected_args[4], [])
        self.assertEqual(expected_kwargs["displacement_scale"], 0)

    @patch("PIL.Image.fromarray")
    @patch("parallax_maker.webui.generate_depth_map")
    @patch("parallax_maker.webui.postprocess_depth_map")
    @patch("parallax_maker.webui.export_gltf")
    def test_export_state_as_gltf_with_displacement(
        self,
        mock_export_gltf,
        mock_postprocess_depth_map,
        mock_generate_depth_map,
        mock_image_fromarray,
    ):
        # Test case 2: Displacement scale is greater than 0
        state = AppState()
        state.image_slices = self.state.image_slices

        mock_export_gltf.return_value = Path("output.gltf")

        mock_generate_depth_map.return_value = np.zeros((100, 100), dtype=np.float32)
        mock_postprocess_depth_map.return_value = np.zeros((100, 100), dtype=np.uint8)

        # path does not exist
        self.mock_depth_file.exists.return_value = False

        # mocking depthmap image saving
        mock_image = MagicMock(spec=Image.Image)
        mock_image_fromarray.return_value = mock_image

        result = export_state_as_gltf(
            self.state, "output_dir", Camera(10, 100, 50), 1, "midas"
        )

        self.assertEqual(result, Path("output.gltf"))
        self.assertEqual(mock_generate_depth_map.call_count, 3)
        self.assertEqual(mock_postprocess_depth_map.call_count, 3)
        self.assertEqual(mock_image_fromarray.call_count, 3)
        mock_image.save.assert_called_with(self.mock_depth_file, compress_level=1)

        # Compare individual elements of card_corners_3d_list
        expected_call = mock_export_gltf.call_args_list[0]
        expected_args, expected_kwargs = expected_call
        self.assertEqual(expected_args[0], Path("output_dir/model.gltf"))

        image_slices_filenames = [slice.filename for slice in self.state.image_slices]
        self.assertEqual(expected_args[3], image_slices_filenames)
        self.assertEqual(expected_args[4], [self.mock_depth_file] * 3)
        self.assertEqual(expected_kwargs["displacement_scale"], 1)

    @patch("parallax_maker.webui.export_gltf")
    def test_export_state_as_gltf_with_upscaled(self, mock_export_gltf):
        # Test case 3: Upscaled slices exist
        state = AppState()
        state.image_slices = self.state.image_slices

        # Pretend the upscaled file exists
        mock_upscaled_file = MagicMock()
        mock_upscaled_file.exists.return_value = True
        self.state.upscaled_filename.return_value = mock_upscaled_file

        result = export_state_as_gltf(
            self.state, "output_dir", Camera(10, 100, 50), 1, "midas"
        )

        # Compare individual elements of card_corners_3d_list
        expected_call = mock_export_gltf.call_args_list[0]
        expected_args, expected_kwargs = expected_call
        self.assertEqual(expected_args[0], Path("output_dir/model.gltf"))
        self.assertEqual(expected_args[3], [mock_upscaled_file] * 3)
        self.assertEqual(expected_args[4], [self.mock_depth_file] * 3)
        self.assertEqual(expected_kwargs["displacement_scale"], 1)

    # TODO: Add more test cases for generate_depth_map, postprocess_depth_map, and export_gltf


class TestSliceUpload(unittest.TestCase):
    @patch("parallax_maker.webui.ctx")
    @patch("parallax_maker.webui.AppState.from_cache")
    def test_filename_none(self, mock_from_cache, mock_ctx):
        with self.assertRaises(PreventUpdate):
            slice_upload(None, None, None)
        mock_from_cache.assert_not_called()

    @patch("parallax_maker.webui.ctx")
    @patch("parallax_maker.webui.AppState.from_cache")
    def test_empty_image_slices(self, mock_from_cache, mock_ctx):
        mock_state = MagicMock(spec=AppState)
        mock_state.image_slices = []
        mock_from_cache.return_value = mock_state

        with self.assertRaises(PreventUpdate):
            slice_upload(None, "appstate-random", None)
        mock_from_cache.assert_called_once_with("appstate-random")

    @patch("parallax_maker.webui.ctx")
    @patch("parallax_maker.webui.AppState.from_cache")
    def test_contents_none(self, mock_from_cache, mock_ctx):
        mock_state = MagicMock(spec=AppState)
        mock_state.image_slices = [np.zeros((100, 100, 4))]
        mock_from_cache.return_value = mock_state
        mock_ctx.triggered_id = {"index": 0}

        with self.assertRaises(PreventUpdate):
            slice_upload([None], "appstate-random", None)
        mock_from_cache.assert_called_once_with("appstate-random")

    @patch("parallax_maker.webui.ctx")
    @patch("parallax_maker.webui.AppState.from_cache")
    @patch("parallax_maker.slice.filename_add_version")
    @patch("parallax_maker.webui.blend_with_alpha")
    @patch("parallax_maker.slice.ImageSlice.save_image")
    def test_valid_upload(
        self,
        mock_imwrite,
        mock_blend,
        mock_filename_add_version,
        mock_from_cache,
        mock_ctx,
    ):
        mock_state = MagicMock(spec=AppState)
        mock_state.image_slices = [
            ImageSlice(np.zeros((100, 100, 4), dtype=np.uint8), filename="slice0.png"),
            ImageSlice(np.ones((100, 100, 4), dtype=np.uint8), filename="slice1.png"),
        ]
        mock_from_cache.return_value = mock_state
        mock_ctx.triggered_id = {"index": 1}
        mock_filename_add_version.return_value = "slice1_v1.png"

        content = to_image_url(np.ones((100, 100, 4), dtype=np.uint8))
        print(content[:50])

        result = slice_upload([None, content], "appstate-random", [])

        mock_imwrite.assert_called()
        self.assertEqual(result[0], True)
        self.assertEqual(len(result[1]), 1)
        self.assertIn(
            "Received image slice upload for slice 1 at slice1_v1.png", result[1][0]
        )
        mock_from_cache.assert_called_once_with("appstate-random")
        mock_filename_add_version.assert_called_once_with("slice1.png")
        mock_blend.assert_called_once()
        mock_state.to_file.assert_called_once_with(
            "appstate-random",
            save_image_slices=False,
            save_depth_map=False,
            save_input_image=False,
        )
        self.assertIsInstance(mock_state.imgData, Image.Image)

    @patch("parallax_maker.webui.ctx")
    @patch("parallax_maker.webui.AppState.from_cache")
    @patch("parallax_maker.slice.filename_add_version")
    @patch("parallax_maker.webui.blend_with_alpha")
    @patch("parallax_maker.slice.ImageSlice.save_image")
    def test_valid_upload_different_ratio(
        self,
        mock_imwrite,
        mock_blend,
        mock_filename_add_version,
        mock_from_cache,
        mock_ctx,
    ):
        mock_state = MagicMock(spec=AppState)
        mock_state.image_slices = [
            ImageSlice(np.zeros((100, 100, 4), dtype=np.uint8), filename="slice0.png"),
            ImageSlice(np.ones((100, 100, 4), dtype=np.uint8), filename="slice1.png"),
        ]
        mock_from_cache.return_value = mock_state
        mock_ctx.triggered_id = {"index": 1}
        mock_filename_add_version.return_value = "slice1_v1.png"

        content = to_image_url(np.ones((110, 99, 4), dtype=np.uint8))

        result = slice_upload([None, content], "appstate-random", [])

        mock_imwrite.assert_called()
        self.assertEqual(result[0], True)
        self.assertEqual(len(result[1]), 2)
        self.assertIn("Fixing aspect ratio from", result[1][0])
        mock_from_cache.assert_called_once_with("appstate-random")
        mock_filename_add_version.assert_called_once_with("slice1.png")
        mock_blend.assert_called_once()
        mock_state.to_file.assert_called_once_with(
            "appstate-random",
            save_image_slices=False,
            save_depth_map=False,
            save_input_image=False,
        )
        self.assertIsInstance(mock_state.imgData, Image.Image)


class TestUpdateSlices(unittest.TestCase):

    def setUp(self):
        # Mocks and AppState setup
        self.filename = "test_file"
        self.mock_state = AppState()

        # Mock the AppState.from_cache method
        self.patcher = patch(
            "parallax_maker.webui.AppState.from_cache", return_value=self.mock_state
        )
        self.mock_from_cache = self.patcher.start()

    def tearDown(self):
        self.patcher.stop()

    def test_no_filename(self):
        ignored_data = MagicMock()
        # Testing when filename is None
        with self.assertRaises(PreventUpdate):
            update_slices(ignored_data, None)

    def test_no_image_slices(self):
        # Test with no image slices
        self.mock_state.image_slices = []
        ignored_data = MagicMock()
        result = update_slices(ignored_data, self.filename)
        self.assertEqual(result, ([], "", no_update))

    def test_no_depth_map_data(self):
        # Test without depthMapData
        self.mock_state.image_slices = [MagicMock()]
        self.mock_state.depthMapData = None
        ignored_data = MagicMock()
        with self.assertRaises(PreventUpdate):
            update_slices(ignored_data, self.filename)

    @patch(
        "parallax_maker.webui.AppState.serve_slice_image_composed",
        return_value="composed_image_data",
    )
    @patch("parallax_maker.webui.AppState.serve_slice_image", return_value="image_data")
    @patch("parallax_maker.slice.ImageSlice.can_undo", return_value=False)
    def test_full_functionality(
        self, mock_serve_slice_image_composed, mock_can_undo, mock_serve_slice_image
    ):
        # Simulate state with several slices
        self.mock_state.image_slices = [
            ImageSlice(depth=0, filename="slice1.png"),
            ImageSlice(depth=1, filename="slice2.png"),
        ]
        self.mock_state.depthMapData = True
        self.mock_state.selected_slice = 0
        self.mock_state.use_checkerboard = True

        ignored_data = MagicMock()
        result = update_slices(ignored_data, self.filename)

        # Check that the result is correctly generated images and the composed image
        img_container, gen_slice_output, img_data = result
        self.assertEqual(len(img_container), 2)
        self.assertEqual(img_data, "composed_image_data")

    @patch("parallax_maker.webui.AppState.serve_slice_image", return_value="image_data")
    @patch("parallax_maker.slice.ImageSlice.can_undo", return_value=False)
    def test_corner_cases(self, mock_can_undo, mock_serve_slice_image):
        # Simulate corner cases where selected_slice is None
        self.mock_state.image_slices = [ImageSlice(depth=0, filename="slice1.png")]
        self.mock_state.depthMapData = True
        self.mock_state.selected_slice = None

        ignored_data = MagicMock()
        result = update_slices(ignored_data, self.filename)

        # Check that the result correctly handles the corner case and does not update the image data
        img_container, gen_slice_output, img_data = result
        self.assertEqual(len(img_container), 1)
        self.assertEqual(img_data, no_update)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# (c) 2024 Niels Provos
#

import argparse
import base64
from io import BytesIO
import numba as nb
import torch
from diffusers import AutoPipelineForInpainting, AutoPipelineForImage2Image
from diffusers.utils import make_image_grid
import cv2
import numpy as np
from PIL import Image, ImageFilter

from .utils import torch_get_device, find_square_bounding_box, feather_mask, timeit
from .automatic1111 import create_img2img_payload, make_img2img_request
from .comfyui import inpainting_comfyui
from .stabilityai import StabilityAI


class InpaintingModel:
    # local models running on the local GPU
    MODELS = [
        "kandinsky-community/kandinsky-2-2-decoder-inpaint",
        "runwayml/stable-diffusion-v1-5",
        "diffusers/stable-diffusion-xl-1.0-inpainting-0.1",
        "stabilityai/stable-diffusion-3-medium-diffusers",
    ]

    # reached via API end points
    EXTERNAL_MODELS = ["automatic1111", "comfyui", "stabilityai"]

    def __init__(
        self, model="diffusers/stable-diffusion-xl-1.0-inpainting-0.1", **kwargs
    ):
        """
        Initializes the Inpainting class.

        Args:
            model (str): The name of the model to be used for inpainting.
                Defaults to "diffusers/stable-diffusion-xl-1.0-inpainting-0.1". Can also
                be "automatic1111" or "comfyui" to use an external API server.
            **kwargs: Additional keyword arguments to be passed to the class.
                At the moment, automatic1111 requires a "server_address" argument and
                ComfyUI requires a "server_address" and "workflow_path" argument.
                StabilityAI requires an "api_key" argument.

        Attributes:
            model (str): The name of the model used for inpainting.
            kwargs (dict): Additional keyword arguments passed to the class.
        """
        assert (
            model in self.MODELS or model in self.EXTERNAL_MODELS
        ), f"Model {model} is not supported."
        self.model = model
        self.variant = ""
        self._dimension = 512
        self._supports_feathering = True
        self.kwargs = kwargs
        self.pipeline = None
        self.server_address = None
        self.workflow_path = None  # use for comfyui

    def __eq__(self, other):
        if not isinstance(other, InpaintingModel):
            return False
        return (
            self.model == other.model
            and self.variant == other.variant
            and self._dimension == other._dimension
        )

    def load_model(self):
        if self.pipeline is not None:
            return self.pipeline

        # diffusers models
        if self.model in self.MODELS:
            return self.load_diffusers_model()

        if self.model in ["automatic1111", "comfyui"]:
            return self.load_external_model()

        if self.model == "stabilityai":
            return self.load_stability_model()

        return None

    def load_diffusers_model(self):
        models = {
            "kandinsky-community/kandinsky-2-2-decoder-inpaint": {
                "dimension": 768,
                "variant": "",
                "pipeline": AutoPipelineForInpainting.from_pretrained,
            },
            "diffusers/stable-diffusion-xl-1.0-inpainting-0.1": {
                "dimension": 1024,
                "variant": "fp16",
                "pipeline": AutoPipelineForInpainting.from_pretrained,
            },
            "stabilityai/stable-diffusion-3-medium-diffusers": {
                "dimension": 1024,
                "variant": "",
                "pipeline": AutoPipelineForImage2Image.from_pretrained,
            },
        }

        assert self.model in models
        model_info = models[self.model]
        self._dimension = model_info["dimension"]
        self.variant = model_info["variant"]

        kwargs = {}
        if self.variant:
            kwargs["variant"] = self.variant

        self.pipeline = model_info["pipeline"](
            self.model, torch_dtype=torch.float16, **kwargs
        )

        device = torch_get_device()
        self.pipeline.to(device)

        return self.pipeline

    def load_external_model(self):
        assert "server_address" in self.kwargs
        self._dimension = 1024
        self.pipeline = self.kwargs["server_address"]
        self.server_address = self.kwargs["server_address"]
        if self.model == "comfyui":
            assert "workflow_path" in self.kwargs
            self.workflow_path = self.kwargs["workflow_path"]
        return self.pipeline

    def load_stability_model(self):
        assert "api_key" in self.kwargs
        self.pipeline = StabilityAI(self.kwargs["api_key"])
        assert self.pipeline.validate_key()
        self._dimension = None  # the model will handle resizing
        return self.pipeline

    def get_dimension(self):
        return self._dimension

    def inpaint(
        self,
        prompt,
        negative_prompt,
        init_image,
        mask_image,
        strength=0.7,
        guidance_scale=8,
        num_inference_steps=50,
        padding=50,
        blur_radius=50,
        crop=False,
        seed=-1,
    ):
        """
        Perform inpainting on an image using the specified model.

        Args:
            prompt (str): The prompt for the inpainting process.
            negative_prompt (str): The negative prompt for the inpainting process.
            init_image (PIL.Image.Image or np.array): The initial image to be inpainted.
            mask_image (PIL.Image.Image or np.array): The mask image indicating the areas to be inpainted.
            strength (float, optional): The strength of the inpainting process. Defaults to 0.7.
            guidance_scale (int, optional): The guidance scale for the inpainting process. Defaults to 8.
            num_inference_steps (int, optional): The number of inference steps for the inpainting process. Defaults to 50.
            padding (int, optional): The padding value for cropping the image. Defaults to 50.
            blur_radius (int, optional): The blur radius for feathering the mask image. Defaults to 50.
            crop (bool, optional): Whether to crop the image after inpainting. Defaults to False.
            seed (int, optional): The seed value for the inpainting process. Defaults to -1.

        Returns:
            PIL.Image.Image: The inpainted image.
        """
        if not isinstance(init_image, Image.Image):
            init_image = Image.fromarray(init_image)
        if not isinstance(mask_image, Image.Image):
            mask_image = Image.fromarray(mask_image)

        # Convert images to RGBA mode
        if init_image.mode != "RGBA":
            init_image = init_image.convert("RGBA")

        original_init_image = init_image.copy()

        if crop:
            bounding_box = find_square_bounding_box(mask_image, padding=padding)
            init_image = init_image.crop(bounding_box)
            mask_image = mask_image.crop(bounding_box)

        if blur_radius > 0 and self._supports_feathering:
            blurred_mask = feather_mask(mask_image, num_expand=blur_radius)
        else:
            blurred_mask = mask_image

        # resize images to match the model input size
        dimension = self.get_dimension()
        if dimension:
            resize_init_image = init_image.convert("RGB").resize((dimension, dimension))
            resize_mask_image = blurred_mask.resize((dimension, dimension))
        else:
            resize_init_image = init_image.convert("RGB")
            resize_mask_image = blurred_mask

        # XXX - refactor this to make it more readable
        if self.model == "stabilityai/stable-diffusion-3-medium-diffusers":
            image = self.inpaint_image2image(
                resize_init_image,
                prompt,
                negative_prompt,
                strength,
                guidance_scale,
                num_inference_steps,
                seed,
            )
        elif self.model in self.MODELS:
            image = self.inpaint_diffusers(
                resize_init_image,
                resize_mask_image,
                prompt,
                negative_prompt,
                strength,
                guidance_scale,
                num_inference_steps,
                seed,
            )
        elif self.model == "automatic1111":
            image = self.inpaint_automatic1111(
                resize_init_image,
                resize_mask_image,
                prompt,
                negative_prompt,
                strength,
                guidance_scale,
                num_inference_steps,
                seed,
            )
        elif self.model == "comfyui":
            image = inpainting_comfyui(
                self.server_address,
                self.workflow_path,
                resize_init_image,
                resize_mask_image,
                prompt,
                negative_prompt,
                strength=strength,
                cfg_scale=guidance_scale,
                steps=num_inference_steps,
                seed=seed,
            )
        elif self.model == "stabilityai":
            image = self.pipeline.image_to_image(
                resize_init_image,
                prompt,
                negative_prompt=negative_prompt,
                strength=strength,
            )
        else:
            raise ValueError(f"Model {self.model} is not supported.")

        image = image.resize(init_image.size, resample=Image.LANCZOS)
        image = image.convert("RGBA")

        image.putalpha(blurred_mask)

        image = Image.alpha_composite(init_image, image)

        if crop:
            original_init_image.paste(image, bounding_box[:2])
            image = original_init_image

        return image

    def inpaint_automatic1111(
        self,
        resize_init_image,
        resize_mask_image,
        prompt,
        negative_prompt,
        strength,
        guidance_scale,
        num_inference_steps,
        seed,
    ):
        server_address = self.kwargs["server_address"]
        payload = create_img2img_payload(
            resize_init_image,
            prompt,
            negative_prompt,
            mask_image=resize_mask_image,
            strength=strength,
            steps=num_inference_steps,
            cfg_scale=guidance_scale,
        )
        images = make_img2img_request(server_address, payload)
        if len(images) != 1:
            raise ValueError("Expected one image from the server.")
        return Image.open(BytesIO(base64.b64decode(images[0])))

    def inpaint_diffusers(
        self,
        resize_init_image,
        resize_mask_image,
        prompt,
        negative_prompt,
        strength,
        guidance_scale,
        num_inference_steps,
        seed,
    ):
        if seed == -1:
            seed = np.random.randint(0, 2**32)
        generator = torch.Generator(torch_get_device()).manual_seed(seed)
        pipeline = self.pipeline
        image = pipeline(
            prompt=prompt,
            negative_prompt=negative_prompt,
            image=resize_init_image,
            mask_image=resize_mask_image,
            generator=generator,
            strength=strength,
            guidance_scale=guidance_scale,
            num_inference_steps=num_inference_steps,
        ).images[0]

        return image

    def inpaint_image2image(
        self,
        resize_init_image,
        prompt,
        negative_prompt,
        strength,
        guidance_scale,
        num_inference_steps,
        seed,
    ):
        result = self.pipeline(
            prompt=prompt,
            negative_prompt=negative_prompt,
            image=resize_init_image,
            strength=strength,
            guidance_scale=guidance_scale,
            num_inference_steps=num_inference_steps,
        )
        return result.images[0]


def create_inpainting_pipeline(model, workflow, state):
    workflow_path = None
    if model == "comfyui":
        if workflow is not None and len(workflow) > 0:
            workflow_path = state.workflow_path()

            need_to_update = False
            if not workflow_path.exists():
                need_to_update = True
            else:
                old_workflow = workflow_path.read_bytes()
                if old_workflow != workflow:
                    need_to_update = True

            if need_to_update:
                # dcc.Upload always has the format 'data:filetype;base64,'
                workflow = workflow.split(",")[1]
                workflow = base64.b64decode(workflow)
                workflow_path.write_bytes(workflow)
                print("ComfyUI workflow updated")

    pipeline = InpaintingModel(
        model,
        server_address=state.server_address,
        workflow_path=workflow_path,
        api_key=state.api_key,
    )
    if state.pipeline_spec is None or state.pipeline_spec != pipeline:
        state.pipeline_spec = pipeline
        pipeline.load_model()
    return state.pipeline_spec


def prefetch_models(fetch_all_models=False):
    """
    Prefetch all models to avoid loading them multiple times.

    Args:
        fetch_all_models (bool): Whether to fetch all models.

    Returns:
        None
    """
    if fetch_all_models:
        for model in InpaintingModel.MODELS:
            InpaintingModel(model).load_model()
    else:
        InpaintingModel().load_model()


def load_image_from_file(image_path, mode="RGB"):
    image = Image.open(image_path)
    image = image.convert(mode)
    return image


@nb.jit(nopython=True, cache=True)
def find_nearest_alpha(alpha, near_a):  # pragma: no cover
    """
    Finds the nearest alpha values in pixel indices for each pixel in the given alpha image.

    Args:
        alpha (numpy.ndarray): The alpha channel represented as a 2D numpy array.
        near_a (numpy.ndarray): The array to store the nearest alpha values.

    Returns:
        None: This function modifies the `nearest_alpha` array in-place.

    Description:
        Uses dynamic programming to find the nearest alpha values for each pixel in the alpha image.

        The `near_a` array is modified in-place to store the nearest alpha values for each pixel.
    """
    height, width = alpha.shape
    for i in range(height):
        for j in range(width):
            if alpha[i, j] == 255:
                near_a[i, j, 0, 0] = i
                near_a[i, j, 0, 1] = j
                near_a[i, j, 1, 0] = i
                near_a[i, j, 1, 1] = j
                near_a[i, j, 2, 0] = i
                near_a[i, j, 2, 1] = j
                near_a[i, j, 3, 0] = i
                near_a[i, j, 3, 1] = j

    for i in range(height):
        for j in range(width):
            if alpha[i, j] != 255:
                if i > 0:
                    near_a[i, j, 0] = near_a[i - 1, j, 0]
                if j > 0:
                    near_a[i, j, 1] = near_a[i, j - 1, 1]
            if alpha[height - i - 1, width - j - 1] != 255:
                if i > 0:
                    near_a[height - i - 1, width - j - 1, 2] = near_a[
                        height - i, width - j - 1, 2
                    ]
                if j > 0:
                    near_a[height - i - 1, width - j - 1, 3] = near_a[
                        height - i - 1, width - j, 3
                    ]


@nb.jit(nopython=True, parallel=True)
def patch_pixels(image, mask_image, nearest_alpha, patch_indices):  # pragma: no cover
    """
    Fills in the missing pixels in the image using a patch-based inpainting algorithm.

    Args:
        image (ndarray): The input image with missing pixels.
        mask_image (ndarray): The mask image indicating the locations of missing pixels.
        nearest_alpha (ndarray): The nearest alpha values for each pixel in the image.
        patch_indices (ndarray): The indices of the pixels to be inpainted.

    Returns:
        None. The input image is modified in-place.

    """
    for idx in nb.prange(len(patch_indices)):
        i, j = patch_indices[idx]
        surrounding_pixels = np.zeros((4, 3))
        distances = np.zeros(4)
        count = 0

        for direction in range(4):
            test_y, test_x = nearest_alpha[i, j, direction]
            if test_y != -1 and test_x != -1:
                distance = max(abs(i - test_y), abs(j - test_x))
                surrounding_pixels[count] = image[test_y, test_x, :3]
                distances[count] = distance
                count += 1

        if count > 0:
            weights = 1 / (distances[:count] + 1)
            average_color = np.sum(
                surrounding_pixels[:count] * weights.reshape(-1, 1), axis=0
            ) / np.sum(weights)
            # alpha blend the patched color with the original color
            alpha = image[i, j, 3]
            orig_color = image[i, j, :3]
            image[i, j, :3] = (
                alpha / 255.0 * orig_color + (255.0 - alpha) / 255.0 * average_color
            )
            image[i, j, 3] = max(mask_image[i, j], image[i, j, 3])


@timeit
def patch_image(image, mask_image):
    """
    Finds the area of the image that has no alpha channel and is overlapped by the mask image.
    It patches those pixels by approximating the color by finding the average color of the nearest pixels that
    have alpha channel.

    Args:
        image (np.array): The image with alpha channel to patch. Modified in place.
        mask_image (np.array): The mask image with a single channel.

    Returns:
        np.array: The patched image with the same shape as the input image.

    Raises:
        ValueError: If the input image does not have an alpha channel.
    """
    if image.shape[2] != 4:
        raise ValueError("The input image must have an alpha channel.")

    height, width, _ = image.shape

    alpha = image[:, :, 3].copy()
    alpha = cv2.GaussianBlur(alpha, (5, 5), 0)

    # Create a boolean mask indicating pixels that need patching
    patch_mask = (alpha < 255) & (mask_image > 0)

    # Find the indices of pixels that need patching
    patch_indices = np.argwhere(patch_mask)

    # Initialize the nearest_alpha array with -1 (indicating no nearest alpha pixel)
    nearest_alpha = np.full((height, width, 4, 2), -1, dtype=int)

    find_nearest_alpha(alpha, nearest_alpha)

    patch_pixels(image, mask_image, nearest_alpha, patch_indices)

    # find the alpha channel of the patched image that is new and not part of the original image
    inverted_alpha = 255 - alpha
    new_alpha = np.minimum(mask_image, inverted_alpha)
    new_alpha = np.expand_dims(new_alpha, axis=2)
    new_alpha = new_alpha / 255.0

    blurred_image = cv2.GaussianBlur(image, (155, 155), 0)

    # composite
    image[:, :, :3] = (
        image[:, :, :3] * (1 - new_alpha) + blurred_image[:, :, :3] * new_alpha
    )

    return image


def main():  # pragma: no cover
    parser = argparse.ArgumentParser(
        description="Inpain images based on a mask and prompt."
    )
    parser.add_argument("-i", "--image", type=str, help="Path to the input image")
    parser.add_argument("-m", "--mask", type=str, help="Path to the mask image")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="inpainted_image.png",
        help="Path to the output filename",
    )
    parser.add_argument(
        "-a",
        "--patch",
        action="store_true",
        help="Patch the image using the fast inpainting algorithm.",
    )
    parser.add_argument(
        "-s",
        "--strength",
        type=float,
        default=0.7,
        help="The strength of diffusion applied to the image",
    )
    parser.add_argument(
        "-p",
        "--prompt",
        type=str,
        default="a black cat with glowing eyes, cute, adorable, disney, pixar, highly detailed, 8k",
        help="The positive prompt to use for inpainting",
    )
    parser.add_argument(
        "-n",
        "--negative-prompt",
        type=str,
        default="bad anatomy, deformed, ugly, disfigured",
        help="The negative prompt to use for inpainting",
    )
    args = parser.parse_args()

    init_image = load_image_from_file(args.image, mode="RGBA")
    mask_image = load_image_from_file(args.mask, mode="L")

    # blur the mask image to feather the edges
    mask_image = mask_image.filter(ImageFilter.GaussianBlur(55))

    init_image = patch_image(np.array(init_image), np.array(mask_image))
    init_image = Image.fromarray(init_image)

    if not args.patch:
        pipeline = InpaintingModel("stabilityai/stable-diffusion-3-medium-diffusers")
        # pipeline = InpaintingModel(
        #    'automatic1111', server_address='localhost:7860')
        # pipeline = InpaintingModel(
        #    'comfyui',
        #    workflow_path='workflow.json',
        #    server_address='localhost:8188')

        pipeline.load_model()

        new_image = pipeline.inpaint(
            args.prompt,
            args.negative_prompt,
            init_image,
            mask_image,
            strength=args.strength,
            crop=True,
        )
    else:
        new_image = init_image

    new_image.save(args.output, format="PNG")

    image = make_image_grid([init_image, mask_image, new_image], rows=1, cols=3)

    image = np.array(image)

    cv2.imshow("Inpainting", cv2.cvtColor(image, cv2.COLOR_RGB2BGR))
    cv2.waitKey(0)


if __name__ == "__main__":
    main()



import argparse
import json
import io
import os
from PIL import Image
from urllib import request, parse
import random
import requests

import uuid
import websocket


def load_workflow(workflow_path):
    try:
        with open(workflow_path, "r") as file:
            workflow = json.load(file)
            return json.dumps(workflow)
    except Exception as e:
        print(f"Failed to load workflow: {e}")
        return None


def patch_inpainting_workflow(
    workflow,
    image,
    mask,
    prompt,
    negative_prompt,
    strength=0.8,
    steps=1,
    cfg_scale=5.0,
    seed=-1,
):
    workflow = json.loads(workflow)

    # find the sampler note in workflow
    sampler_id = [
        key for key, value in workflow.items() if value["class_type"] == "KSampler"
    ][0]

    sampler = workflow.get(sampler_id)
    sampler["inputs"]["denoise"] = strength
    sampler["inputs"]["steps"] = steps
    sampler["inputs"]["cfg"] = cfg_scale

    if seed == -1:
        seed = random.randint(0, 2**32 - 1)
    sampler["inputs"]["seed"] = seed

    positive_id = sampler["inputs"]["positive"][0]
    positive_node = workflow.get(positive_id)
    positive_node["inputs"]["text"] = prompt

    negative_id = sampler["inputs"]["negative"][0]
    negative_node = workflow.get(negative_id)
    negative_node["inputs"]["text"] = negative_prompt

    latent_id = sampler["inputs"]["latent_image"][0]
    latent_node = workflow.get(latent_id)

    # this is very fragile and is not going to work in general
    mask_id = None
    if latent_node["class_type"] == "SetLatentNoiseMask":
        mask_id = latent_node["inputs"]["mask"][0]
        latent_id = latent_node["inputs"]["samples"][0]
        latent_node = workflow.get(latent_id)

    if latent_node["class_type"] == "VAEEncode":
        image_id = latent_node["inputs"]["pixels"][0]
        if mask_id is None:
            mask_id = latent_node["inputs"]["mask"][0]
    else:
        raise ValueError(f"Unknown node type: {latent_node['class_type']}")

    image_node = workflow.get(image_id)
    mask_node = workflow.get(mask_id)

    assert "image" in image_node["inputs"], "Image node does not have an image input"
    assert "image" in mask_node["inputs"], "Mask node does not have an image input"

    image_node["inputs"]["image"] = image
    mask_node["inputs"]["image"] = mask

    return workflow


def queue_prompt(server_address, client_id, workflow):
    # send the workflow to the server
    p = {"prompt": workflow, "client_id": client_id}
    data = json.dumps(p).encode("utf-8")
    req = request.Request(f"http://{server_address}/prompt", data=data)
    return json.loads(request.urlopen(req).read())


def upload_image(server_address, file, overwrite=False):
    try:
        files = {
            "image": open(file, "rb"),
        }
        data = {"type": "input", "overwrite": str(overwrite).lower()}

        # Making the POST request
        response = requests.post(
            f"http://{server_address}/upload/image", files=files, data=data, timeout=500
        )

        if response.status_code == 200:
            data = response.json()
            path = data["name"]
            return path
        else:
            print(f"HTTP Error: {response.status_code} - {response.reason}")
            return None

    except Exception as error:
        print(f"An error occurred: {error}")
        return None


def get_image(server_address, filename, subfolder, folder_type):
    data = {"filename": filename, "subfolder": subfolder, "type": folder_type}
    url_values = parse.urlencode(data)
    with request.urlopen(f"http://{server_address}/view?{url_values}") as response:
        return response.read()


def get_images(server_address, client_id, prompt_id):
    ws = websocket.WebSocket()
    ws.connect(f"ws://{server_address}/ws?clientId={client_id}")

    output_images = {}
    while True:
        out = ws.recv()
        if isinstance(out, str):
            message = json.loads(out)
            if message["type"] == "executing":
                data = message["data"]
                if data["node"] is None and data["prompt_id"] == prompt_id:
                    break  # Execution is done
        else:
            continue  # previews are binary data

    history = get_history(server_address, prompt_id)[prompt_id]
    for o in history["outputs"]:
        for node_id in history["outputs"]:
            node_output = history["outputs"][node_id]
            if "images" in node_output:
                images_output = []
                for image in node_output["images"]:
                    image_data = get_image(
                        server_address,
                        image["filename"],
                        image["subfolder"],
                        image["type"],
                    )
                    images_output.append(image_data)
            output_images[node_id] = images_output

    return output_images


def get_history(server_address, prompt_id):
    with request.urlopen(f"http://{server_address}/history/{prompt_id}") as response:
        return json.loads(response.read())


def temporary_filename(prefix="tmp", suffix=".png"):
    return f"{prefix}-{uuid.uuid4()}{suffix}"


def inpainting_comfyui(
    server_address,
    workflow_path,
    image,
    mask,
    prompt,
    negative_prompt,
    strength=0.5,
    steps=40,
    cfg_scale=5.0,
    seed=-1,
):
    """
    Perform inpainting using an external ComfyUI server.

    Args:
        server_address (str): The address of the server.
        workflow_path (str): The path to the workflow file.
        image (PIL.Image): The input image.
        mask (PIL.Image): The mask indicating the areas to be inpainted.
        prompt (str): The prompt for the inpainting process.
        negative_prompt (str): The negative prompt for the inpainting process.
        strength (float, optional): The strength of the inpainting. Defaults to 0.5.
        steps (int, optional): The number of steps for the inpainting process. Defaults to 40.
        cfg_scale (float, optional): The scale factor for the configuration. Defaults to 5.0.
        seed (int, optional): The seed for the inpainting process. Defaults to -1.

    Returns:
        PIL.Image: The inpainted image.
    """
    workflow = load_workflow(workflow_path)
    if workflow is None:
        return None

    client_id = str(uuid.uuid4())

    # we assume PIL images
    image_path = temporary_filename()
    image.save(image_path)
    mask_path = temporary_filename()
    mask.save(mask_path)

    image = None
    try:
        image = upload_image(server_address, image_path, overwrite=True)
        mask = upload_image(server_address, mask_path, overwrite=True)

        workflow = patch_inpainting_workflow(
            workflow,
            image,
            mask,
            prompt,
            negative_prompt,
            strength=strength,
            steps=steps,
            cfg_scale=cfg_scale,
            seed=seed,
        )

        status = queue_prompt(server_address, client_id, workflow)
        prompt_id = status["prompt_id"]

        images = get_images(server_address, client_id, prompt_id)

        # XXX - only one image for now
        id = list(images.keys())[0]
        image_data = images[id][0]
        image = Image.open(io.BytesIO(image_data))
    finally:
        # remove temporary files
        os.remove(image_path)
        os.remove(mask_path)

    return image


def main():
    argsparse = argparse.ArgumentParser()
    argsparse.add_argument(
        "-w", "--workflow", type=str, help="Path to the workflow file"
    )
    argsparse.add_argument("-i", "--image", type=str, help="Path to the image file")
    argsparse.add_argument("-m", "--mask", type=str, help="Path to the mask file")
    argsparse.add_argument(
        "-p",
        "--prompt",
        type=str,
        default="A beautiful landscape with a mountain in the background",
        help="Positive prompt",
    )
    argsparse.add_argument(
        "-n",
        "--negative-prompt",
        type=str,
        default="out of focus",
        help="Negative prompt",
    )
    argsparse.add_argument("-s", "--strength", default=0.8, type=float, help="Strength")
    argsparse.add_argument(
        "-e", "--steps", type=int, default=40, help="Number of steps for the sampler"
    )
    args = argsparse.parse_args()

    server_address = "localhost:8188"

    image = Image.open(args.image)
    mask = Image.open(args.mask)

    image = inpainting_comfyui(
        server_address,
        args.workflow,
        image,
        mask,
        args.prompt,
        args.negative_prompt,
        args.strength,
    )
    image.show()


if __name__ == "__main__":
    main()
